# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from collections import ChainMap
from typing import Callable, Iterable, TypeVar

from gt4py import eve
from gt4py.eve import utils as eve_utils
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.iterator.transforms import inline_lambdas


@dataclasses.dataclass(frozen=True)
class CannonicalizeBoundSymbolNames(eve.NodeTranslator):
    """
    Given an iterator expression cannonicalize all bound symbol names.

    If two such expression are in the same scope and equal so are their values.

    >>> testee1 = im.lambda_("a")(im.plus("a", "b"))
    >>> cannonicalized_testee1 = CannonicalizeBoundSymbolNames.apply(testee1)
    >>> str(cannonicalized_testee1)
    'λ(_csym_1) → _csym_1 + b'

    >>> testee2 = im.lambda_("c")(im.plus("c", "b"))
    >>> cannonicalized_testee2 = CannonicalizeBoundSymbolNames.apply(testee2)
    >>> assert cannonicalized_testee1 == cannonicalized_testee2
    """

    _uids: eve_utils.UIDGenerator = dataclasses.field(
        init=False, repr=False, default_factory=lambda: eve_utils.UIDGenerator(prefix="_csym")
    )

    @classmethod
    def apply(cls, node: itir.Expr) -> itir.Expr:
        return cls().visit(node, sym_map=ChainMap({}))

    def visit_Lambda(self, node: itir.Lambda, *, sym_map: ChainMap) -> itir.Lambda:
        sym_map = sym_map.new_child()
        for param in node.params:
            sym_map[str(param.id)] = self._uids.sequential_id()

        return im.lambda_(*sym_map.values())(self.visit(node.expr, sym_map=sym_map))

    def visit_SymRef(self, node: itir.SymRef, *, sym_map: dict[str, str]) -> itir.SymRef:
        return im.ref(sym_map[node.id]) if node.id in sym_map else node


def is_equal(a: itir.Expr, b: itir.Expr) -> bool:
    """
    Return true if two expressions have provably equal values.

    Be aware that this function might return false even though the two expression have the same
    value.

    >>> testee1 = im.lambda_("a")(im.plus("a", "b"))
    >>> testee2 = im.lambda_("c")(im.plus("c", "b"))
    >>> assert is_equal(testee1, testee2)

    >>> testee1 = im.lambda_("a")(im.plus("a", "b"))
    >>> testee2 = im.lambda_("c")(im.plus("c", "d"))
    >>> assert not is_equal(testee1, testee2)
    """
    # TODO(tehrengruber): Extend this function cover more cases than just those with equal
    #  structure, e.g., by also canonicalization of the structure.
    return a == b or (
        CannonicalizeBoundSymbolNames.apply(a) == CannonicalizeBoundSymbolNames.apply(b)
    )


def canonicalize_as_fieldop(expr: itir.FunCall) -> itir.FunCall:
    """
    Canonicalize applied `as_fieldop`s.

    In case the stencil argument is a `deref` wrap it into a lambda such that we have a unified
    format to work with (e.g. each parameter has a name without the need to special case).
    """
    assert cpm.is_applied_as_fieldop(expr)

    stencil = expr.fun.args[0]
    domain = expr.fun.args[1] if len(expr.fun.args) > 1 else None
    if cpm.is_ref_to(stencil, "deref"):
        stencil = im.lambda_("arg")(im.deref("arg"))
        new_expr = im.as_fieldop(stencil, domain)(*expr.args)

        return new_expr

    return expr


def _remove_let_alias(let_expr: itir.FunCall) -> itir.FunCall:
    assert cpm.is_let(let_expr)
    is_aliased_let = True
    for param, arg in zip(let_expr.fun.params, let_expr.args, strict=True):
        is_aliased_let &= cpm.is_ref_to(arg, param.id)
    if is_aliased_let:
        assert isinstance(let_expr.fun.expr, itir.FunCall)
        return let_expr.fun.expr
    return let_expr


def unwrap_scan(
    stencil: itir.Lambda | itir.FunCall,
) -> tuple[itir.Lambda, Callable[[itir.Lambda], itir.FunCall | itir.Lambda]]:
    """
    If given a scan, extract stencil part of its scan pass and a back-transformation into a scan.

    If a regular stencil is given the stencil is left as-is and the back-transformation is the
    identity function. This function allows treating a scan stencil like a regular stencil during
    a transformation avoiding the complexity introduced by the different IR format.

    >>> scan = im.call("scan")(
    ...     im.lambda_("state", "arg")(im.plus("state", im.deref("arg"))), True, 0.0
    ... )
    >>> stencil, back_trafo = unwrap_scan(scan)
    >>> str(stencil)
    'λ(arg) → state + ·arg'
    >>> str(back_trafo(stencil))
    'scan(λ(state, arg) → state + ·arg, True, 0.0)'

    In case a regular stencil is given it is returned as-is:

    >>> deref_stencil = im.lambda_("it")(im.deref("it"))
    >>> stencil, back_trafo = unwrap_scan(deref_stencil)
    >>> assert stencil == deref_stencil
    """
    if cpm.is_call_to(stencil, "scan"):
        scan_pass, direction, init = stencil.args
        assert isinstance(scan_pass, itir.Lambda)
        # remove scan pass state to be used by caller
        state_param = scan_pass.params[0]
        stencil_like = im.lambda_(*scan_pass.params[1:])(scan_pass.expr)

        def restore_scan(transformed_stencil_like: itir.Lambda):
            new_scan_pass = im.lambda_(state_param, *transformed_stencil_like.params)(
                _remove_let_alias(
                    im.call(transformed_stencil_like)(
                        *(param.id for param in transformed_stencil_like.params)
                    )
                )
            )
            return im.call("scan")(new_scan_pass, direction, init)

        return stencil_like, restore_scan

    assert isinstance(stencil, itir.Lambda)
    return stencil, lambda s: s


def with_altered_arg(node: itir.FunCall, arg_idx: int, new_arg: itir.Expr | str) -> itir.FunCall:
    """Given a itir.FunCall return a new call with one of its argument replaced."""
    return im.call(node.fun)(
        *(arg if i != arg_idx else im.ensure_expr(new_arg) for i, arg in enumerate(node.args))
    )


def extract_projector(
    node: itir.Expr, cur_projector=None, _depth=0
) -> tuple[itir.Lambda | None, itir.Expr]:
    """
    Extract the projector from an expression (only useful for `scan`s).

    A projector is an expression that consists only of `make_tuple` of `tuple_get` of the same expression,
    possibly in a let statement.

    This is needed for expressions like `as_fieldop(scan(λ(state, val) → {val, state[0]+val}))(inp)[1]`,
    where only element 1 of the state is used. In this example the projector is `λ(_proj) → _proj[1]`.

    Returns the projector and the expression it is applied to.

    Note: Supports only unary projectors. Extend to multi-parameter projectors if needed.
    """
    projector: itir.Lambda | None = None
    expr = node
    if cpm.is_let(node) and len(node.fun.params) == 1:
        # a single param let, it's a projector if the let value aka `node.fun.expr` is a projector
        # > let val = expr
        # >  val[x]
        # > end
        # ->
        # `λ(val) → val[x]`, `expr`
        is_projector, _ = extract_projector(node.fun.expr)
        if is_projector is not None:
            # we can directly use this as projector
            projector = node.fun
            expr = node.args[0]
        else:
            projector = None
            expr = node
    elif cpm.is_call_to(node, "tuple_get"):
        # `expr[x]` -> `λ(_proj) → _proj[x]`, `expr`
        index = node.args[0]
        assert isinstance(index, itir.Literal), index
        projector = im.lambda_(f"_proj{_depth}")(im.tuple_get(index, f"_proj{_depth}"))
        expr = node.args[1]
    elif cpm.is_call_to(node, "make_tuple"):
        # `make_tuple(expr[x0], expr[x1], ...)` -> `λ(_proj) → {_proj[x0], _proj[x1], ...}`, `expr`
        projectors, exprs = zip(*(extract_projector(arg) for arg in node.args))
        if all(p is not None for p in projectors) and all(e == exprs[0] for e in exprs):
            projector = im.lambda_(f"_proj{_depth}")(
                im.call("make_tuple")(*(im.call(p)(im.ref(f"_proj{_depth}")) for p in projectors))
            )
            expr = exprs[0]

    if projector is None:
        return cur_projector, expr
    else:
        # nested projectors, e.g. `expr[x][y]`
        projector = projector if cur_projector is None else im.compose(cur_projector, projector)
        projector = inline_lambdas.InlineLambdas.apply(projector)
        return extract_projector(expr, projector, _depth + 1)


def grid_type_from_domain(domain: itir.FunCall) -> common.GridType:
    if cpm.is_call_to(domain, "cartesian_domain"):
        return common.GridType.CARTESIAN
    else:
        assert cpm.is_call_to(domain, "unstructured_domain")
        return common.GridType.UNSTRUCTURED


def grid_type_from_program(program: itir.Program) -> common.GridType:
    domains = program.walk_values().if_isinstance(itir.SetAt).getattr("domain").to_set()
    grid_types = {grid_type_from_domain(d) for d in domains}
    if len(grid_types) != 1:
        raise ValueError(
            f"Found 'set_at' with more than one 'GridType': '{grid_types}'. This is currently not supported."
        )
    return grid_types.pop()


SymOrStr = TypeVar("SymOrStr", itir.Sym, str)


def unique_symbol(sym: SymOrStr, reserved_names: Iterable[str]) -> SymOrStr:
    """
    Give a symbol and a list of reserved names return a unique symbol with similar or equal name.
    """
    if isinstance(sym, itir.Sym):
        return im.sym(unique_symbol(sym.id, reserved_names), sym.type)  # type: ignore[return-value]  # mypy not smart enough

    assert isinstance(sym, str)
    name: str = sym
    while name in reserved_names:
        name = name + "_"

    return name



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Functions for turning an SDFG into a GPU SDFG."""

from __future__ import annotations

import copy
from typing import Any, Callable, Final, Optional, Sequence, Union

import dace
from dace import (
    dtypes as dace_dtypes,
    properties as dace_properties,
    transformation as dace_transformation,
)
from dace.codegen.targets import cpp as dace_cpp
from dace.sdfg import memlet_utils as dace_mutils, nodes as dace_nodes

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


def gt_gpu_transformation(
    sdfg: dace.SDFG,
    try_removing_trivial_maps: bool = True,
    use_gpu_storage: bool = True,
    gpu_block_size: Optional[Sequence[int | str] | str] = None,
    gpu_launch_bounds: Optional[int | str] = None,
    gpu_launch_factor: Optional[int] = None,
    validate: bool = True,
    validate_all: bool = False,
    **kwargs: Any,
) -> dace.SDFG:
    """Transform an SDFG into a GPU SDFG.

    The transformation expects a rather optimized SDFG and turn it into an SDFG
    capable of running on the GPU.
    The function performs the following steps:
    - If requested, modify the storage location of the non transient arrays such
        that they reside in GPU memory.
    - Call the normal GPU transform function followed by simplify.
    - If requested try to remove trivial kernels.
    - If specified, set the `gpu_block_size` parameters of the Maps to the given value.

    Args:
        sdfg: The SDFG that should be processed.
        try_removing_trivial_maps: Try to get rid of trivial maps by incorporating them.
        use_gpu_storage: Assume that the non global memory is already on the GPU. This
            will avoid the data copy from host to GPU memory.
        gpu_block_size: The size of a thread block on the GPU.
        gpu_launch_bounds: Use this value as `__launch_bounds__` for _all_ GPU Maps.
            Will only take effect if `gpu_block_size` is specified.
        gpu_launch_factor: Use the number of threads times this value as `__launch_bounds__`
            Will only take effect if `gpu_block_size` is specified.
        validate: Perform validation during the steps.
        validate_all: Perform extensive validation.

    Notes:
        - In addition it might fuse Maps together that should not be fused. To prevent
            that you should set `try_removing_trivial_maps` to `False`.
        - The function assumes that the iteration order has been set correctly before
            it is called.

    Todo:
        - Currently only one block size for all maps is given, add more options.
        - Investigate if the order of iteration is not changed (it should not).
        - Investigate if the trivial GPU map remover is still needed.
    """
    assert len(kwargs) == 0, (
        f"gt_gpu_transformation(): found unknown arguments: {', '.join(arg for arg in kwargs.keys())}"
    )

    # Turn all global arrays (which we identify as input) into GPU memory.
    #  This way the GPU transformation will not create this copying stuff.
    if use_gpu_storage:
        for desc in sdfg.arrays.values():
            if isinstance(desc, dace.data.Array) and not desc.transient:
                desc.storage = dace.dtypes.StorageType.GPU_Global

    # Now turn it into a GPU SDFG
    sdfg.apply_gpu_transformations(
        validate=False,
        validate_all=validate_all,
        simplify=False,
    )

    # The documentation recommends to run simplify afterwards
    gtx_transformations.gt_simplify(sdfg, validate=False, validate_all=validate_all)

    if try_removing_trivial_maps:
        # TODO(phimuell): Figuring out if it is still important/needed to do or if
        #   it can be removed, it should definitely be reworked.
        gt_remove_trivial_gpu_maps(
            sdfg=sdfg,
            validate=False,
            validate_all=validate_all,
        )
        gtx_transformations.gt_simplify(sdfg, validate=False, validate_all=validate_all)

    # TODO(phimuell): Fixing the stride problem in DaCe.
    sdfg = gt_gpu_transform_non_standard_memlet(
        sdfg=sdfg,
        map_postprocess=True,
        validate=False,
        validate_all=validate_all,
    )

    # Set the GPU block size if it is known.
    if gpu_block_size is not None:
        gt_set_gpu_blocksize(
            sdfg=sdfg,
            block_size=gpu_block_size,
            launch_bounds=gpu_launch_bounds,
            launch_factor=gpu_launch_factor,
            validate=False,
            validate_all=validate_all,
        )

    if validate_all or validate:
        sdfg.validate()

    return sdfg


def gt_gpu_transform_non_standard_memlet(
    sdfg: dace.SDFG,
    map_postprocess: bool,
    validate: bool = True,
    validate_all: bool = False,
) -> dace.SDFG:
    """Transform some non-standard Melets to Maps.

    The GPU code generator is not able to handle certain sets of Memlets. To
    handle them, the code generator transforms them into copy Maps. The main
    issue is that this transformation happens after the auto optimizer, thus
    the copy-Maps will most likely have the wrong iteration order.

    This function allows to perform the preprocessing step before the actual
    code generation. The function will perform the expansion. If
    `map_postprocess` is `True` then the function will also apply MapFusion,
    to these newly created copy-Maps and set their iteration order correctly.

    A user should not call this function directly, instead this function is
    called by the `gt_gpu_transformation()` function.

    Args:
        sdfg: The SDFG that we process.
        map_postprocess: Enable post processing of the maps that are created.
            See the Note section below.
        validate: Perform validation at the end of the function.
        validate_all: Perform validation also on intermediate steps.

    Note:
        - Currently the function applies some crude heuristic to determine the
            correct loop order.
        - This function should be called after `gt_set_iteration_order()` has run.
    """

    # Expand all non-standard memlets and get the new MapEntries.
    new_maps: set[dace_nodes.MapEntry] = _gt_expand_non_standard_memlets(sdfg)

    # If there are no Memlets that are translated to copy-Maps, then we have nothing to do.
    if len(new_maps) == 0:
        return sdfg

    # This function allows to restrict any fusion operation to the maps
    #  that we have just created.
    def restrict_fusion_to_newly_created_maps_vertical(
        self: gtx_transformations.MapFusionVertical,
        map_exit_1: dace_nodes.MapExit,
        map_entry_2: dace_nodes.MapEntry,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
    ) -> bool:
        return (map_entry_2 in new_maps) or (graph.entry_node(map_exit_1) in new_maps)

    def restrict_fusion_to_newly_created_maps_horizontal(
        self: gtx_transformations.MapFusionHorizontal,
        map_entry_1: dace_nodes.MapEntry,
        map_entry_2: dace_nodes.MapEntry,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
    ) -> bool:
        return (map_entry_1 in new_maps) or (map_entry_2 in new_maps)

    # Now try to fuse the maps together, but restrict them that at least one map
    #  needs to be new.
    # TODO(phimuell): Improve this by replacing it by an explicit loop.
    sdfg.apply_transformations_repeated(
        [
            gtx_transformations.MapFusionVertical(
                only_toplevel_maps=True,
                check_fusion_callback=restrict_fusion_to_newly_created_maps_vertical,
            ),
            gtx_transformations.MapFusionHorizontal(
                only_toplevel_maps=True,
                check_fusion_callback=restrict_fusion_to_newly_created_maps_horizontal,
            ),
        ],
        validate=False,
        validate_all=validate_all,
    )

    # This is a gross hack, but it is needed, for the following reasons:
    #  - The transients have C order while the non-transients have (most
    #       likely) FORTRAN order. So there is not an unique stride dimension.
    #  - The newly created maps have names that does not reflect GT4Py dimensions,
    #       thus we can not use `gt_set_iteration_order()`.
    #  For these reasons we do the simplest thing, which is assuming that the maps
    #  are created in C order and we must make them in FORTRAN order, which means
    #  just swapping the order of the map parameters.
    #  We further assume here, that we only have to process the maps that we have
    #  newly created.
    # NOTE: We can stop relying on this once [PR#1913](https://github.com/GridTools/gt4py/pull/1913)
    #   Has been merged, which is currently blocked by a DaCe PR that has not been
    #   merged.

    maps_to_modify: set[dace_nodes.MapEntry] = set()
    for nsdfg in sdfg.all_sdfgs_recursive():
        for state in nsdfg.states():
            for map_entry in state.nodes():
                if not isinstance(map_entry, dace_nodes.MapEntry):
                    continue
                if map_entry in new_maps:
                    maps_to_modify.add(map_entry)

    # We did not found any of the newly created Map. Thus we **hope** that all new
    #  Maps have been integrated into other Maps, that have the correct names.
    #  But as written above, this is a gross hack!
    if len(maps_to_modify) == 0:
        return sdfg

    for me_to_modify in maps_to_modify:
        map_to_modify: dace_nodes.Map = me_to_modify.map
        map_to_modify.params = list(reversed(map_to_modify.params))
        map_to_modify.range = dace.subsets.Range(
            (r1, r2, r3, t)
            for (r1, r2, r3), t in zip(
                reversed(map_to_modify.range.ranges), reversed(map_to_modify.range.tile_sizes)
            )
        )

    if validate or validate_all:
        sdfg.validate()

    return sdfg


def _gt_expand_non_standard_memlets(
    sdfg: dace.SDFG,
) -> set[dace_nodes.MapEntry]:
    """Finds all non-standard Memlet in the SDFG and expand them.

    The function is used by `gt_gpu_transform_non_standard_memlet()` and performs
    the actual expansion of the Memlet, i.e. turning all Memlets that can not be
    expressed as a `memcpy()` into a Map, copy kernel.
    The function will return the MapEntries of all expanded.

    The function will process the SDFG recursively.
    """
    new_maps: set[dace_nodes.MapEntry] = set()
    for nsdfg in sdfg.all_sdfgs_recursive():
        new_maps.update(_gt_expand_non_standard_memlets_sdfg(nsdfg))
    return new_maps


def _gt_expand_non_standard_memlets_sdfg(
    sdfg: dace.SDFG,
) -> set[dace_nodes.MapEntry]:
    """Implementation of `_gt_expand_non_standard_memlets()` that process a single SDFG."""
    new_maps: set[dace_nodes.MapEntry] = set()
    # The implementation is based on DaCe's code generator, see `dace/codegen/targets/cuda.py`
    #  in the function `preprocess()`
    # NOTE: This implementation needs a DaCe version that includes https://github.com/spcl/dace/pull/2033
    for state in sdfg.states():
        for e in state.edges():
            # We are only interested in edges that connects two access nodes of GPU memory.
            #  However, we must exclude Memlets that are empty.
            if not (
                (not e.data.is_empty())
                and isinstance(e.src, dace_nodes.AccessNode)
                and isinstance(e.dst, dace_nodes.AccessNode)
                and e.src.desc(sdfg).storage == dace_dtypes.StorageType.GPU_Global
                and e.dst.desc(sdfg).storage == dace_dtypes.StorageType.GPU_Global
            ):
                continue

            a: dace_nodes.AccessNode = e.src
            b: dace_nodes.AccessNode = e.dst
            copy_shape, src_strides, dst_strides, _, _ = dace_cpp.memlet_copy_to_absolute_strides(
                None, sdfg, state, e, a, b
            )
            dims = len(copy_shape)
            if dims == 1:
                continue
            elif dims == 2:
                is_fortran_order = src_strides[0] == 1 and dst_strides[0] == 1
                is_c_order = src_strides[-1] == 1 and dst_strides[-1] == 1
                if is_c_order or is_fortran_order:
                    continue

                # NOTE: Special case of continuous copy
                # Example: dcol[0:I, 0:J, k] -> datacol[0:I, 0:J]
                # with copy shape [I, J] and strides [J*K, K], [J, 1]
                if src_strides[-1] != 1 or dst_strides[-1] != 1:
                    try:
                        is_src_cont = src_strides[0] / src_strides[1] == copy_shape[1]
                        is_dst_cont = dst_strides[0] / dst_strides[1] == copy_shape[1]
                    except (TypeError, ValueError):
                        is_src_cont = False
                        is_dst_cont = False
                    if is_src_cont and is_dst_cont:
                        continue

            elif dims > 2:
                if not (src_strides[-1] != 1 or dst_strides[-1] != 1):
                    continue

            # For identifying the new map, we first store all neighbors of `a`.
            old_neighbors_of_a: list[dace_nodes.AccessNode] = [
                edge.dst for edge in state.out_edges(a)
            ]

            if not dace_mutils.can_memlet_be_turned_into_a_map(
                edge=e, state=state, sdfg=sdfg, ignore_strides=True
            ):
                # NOTE: In DaCe, they simply ignore that case and continue to the
                #   code generator. In GT4Py we generate an error.
                raise RuntimeError(f"Unable to turn the not supported edge '{e}' into a copy Map.")

            # Turn the not supported Memlet into a copy Map. We have to do it here,
            #  such that we can then set their iteration order correctly.
            dace_mutils.memlet_to_map(
                edge=e,
                state=state,
                sdfg=sdfg,
                ignore_strides=True,
            )

            # We find the new map by comparing the new neighborhood of `a` with the old one.
            new_nodes: set[dace_nodes.MapEntry] = {
                edge.dst for edge in state.out_edges(a) if edge.dst not in old_neighbors_of_a
            }
            assert any(isinstance(new_node, dace_nodes.MapEntry) for new_node in new_nodes)
            new_maps.update(new_nodes)
    return new_maps


def gt_set_gpu_blocksize(
    sdfg: dace.SDFG,
    block_size: Optional[Sequence[int | str] | str],
    launch_bounds: Optional[int | str] = None,
    launch_factor: Optional[int] = None,
    validate: bool = True,
    validate_all: bool = False,
    **kwargs: Any,
) -> int:
    """Set the block size related properties of _all_ Maps.

    It supports the same arguments as `GPUSetBlockSize`, however it also has
    versions without `_Xd`, these are used as default for the other maps.
    If a version with `_Xd` is specified then it takes precedence.

    Args:
        sdfg: The SDFG to process.
        block_size: The size of a thread block on the GPU.
        launch_bounds: The value for the launch bound that should be used.
        launch_factor: If no `launch_bounds` was given use the number of threads
            in a block multiplied by this number.
        validate: Perform validation at the end of the function.
        validate_all: Perform validation also on intermediate steps.

    Note:
        If a Map is found whose range is smaller than the specified block size for
        that dimension then the block size for that map is reduced.
    """
    for dim in [1, 2, 3]:
        for arg, val in {
            "block_size": block_size,
            "launch_bounds": launch_bounds,
            "launch_factor": launch_factor,
        }.items():
            if f"{arg}_{dim}d" not in kwargs:
                kwargs[f"{arg}_{dim}d"] = val

    setter = GPUSetBlockSize(**kwargs)

    configured_maps = 0
    for state in sdfg.states():
        scope_dict: Union[None, dict[Any, Any]] = None
        cfg_id = state.parent_graph.cfg_id
        state_id = state.block_id
        for node in state.nodes():
            if not isinstance(node, dace_nodes.MapEntry):
                continue
            if scope_dict is None:
                scope_dict = state.scope_dict()
            if scope_dict[node] is not None:
                continue
            candidate = {GPUSetBlockSize.map_entry: node}
            setter.setup_match(
                sdfg=sdfg,
                cfg_id=cfg_id,
                state_id=state_id,
                subgraph=candidate,
                expr_index=0,
                override=True,
            )
            if setter.can_be_applied(state, 0, sdfg, False):
                setter.apply(state, sdfg)
                if validate_all:
                    sdfg.validate()
                configured_maps += 1

    if validate and (not validate_all):
        sdfg.validate()

    return configured_maps


def _make_gpu_block_parser_for(
    dim: int,
) -> Callable[["GPUSetBlockSize", Any], None]:
    """Generates a parser for GPU blocks for dimension `dim`.

    The returned function can be used as parser for the `GPUSetBlockSize.block_size_*d`
    properties.
    """

    def _gpu_block_parser(
        self: GPUSetBlockSize,
        val: Any,
    ) -> None:
        """Used by the setter of `GPUSetBlockSize.block_size`."""
        org_val = val
        if isinstance(val, (tuple | list)):
            pass
        elif isinstance(val, str):
            val = tuple(x.strip() for x in val.split(","))
        elif isinstance(val, int):
            val = (val,)
        else:
            raise TypeError(
                f"Does not know how to transform '{type(org_val).__name__}' into a proper GPU block size."
            )
        if len(val) < dim:
            raise ValueError(
                f"The passed block size only covers {len(val)} dimensions, but dimension was {dim}."
            )
        if 0 < len(val) <= 3:
            val = [*val, *([1] * (3 - len(val)))]
        else:
            raise ValueError(f"Can not parse block size '{org_val}': wrong length")
        try:
            val = [int(x) for x in val]
        except ValueError:
            raise TypeError(
                f"Currently only block sizes convertible to int are supported, you passed '{val}'."
            ) from None

        # Remove over specification.
        for i in range(dim, 3):
            val[i] = 1
        setattr(self, f"_block_size_{dim}d", tuple(val))

    return _gpu_block_parser


def _make_gpu_block_getter_for(
    dim: int,
) -> Callable[["GPUSetBlockSize"], tuple[int, int, int]]:
    """Makes the getter for the block size of dimension `dim`."""

    def _gpu_block_getter(
        self: "GPUSetBlockSize",
    ) -> tuple[int, int, int]:
        """Used as getter in the `GPUSetBlockSize.block_size` property."""
        return getattr(self, f"_block_size_{dim}d")

    return _gpu_block_getter


def _gpu_launch_bound_parser(
    block_size: tuple[int, int, int],
    launch_bounds: int | str | None,
    launch_factor: int | None = None,
) -> str | None:
    """Used by the `GPUSetBlockSize.__init__()` method to parse the launch bounds."""
    if launch_bounds is None and launch_factor is None:
        return None
    elif launch_bounds is None and launch_factor is not None:
        return str(int(launch_factor) * block_size[0] * block_size[1] * block_size[2])
    elif launch_bounds is not None and launch_factor is None:
        assert isinstance(launch_bounds, (str, int))
        return str(launch_bounds)
    else:
        raise ValueError("Specified both `launch_bounds` and `launch_factor`.")


@dace_properties.make_properties
class GPUSetBlockSize(dace_transformation.SingleStateTransformation):
    """Sets the GPU block size on GPU Maps.

    The `block_size` is either a sequence, of up to three integers or a string
    of up to three numbers, separated by comma (`,`). The first number is the size
    of the block in `x` direction, the second for the `y` direction and the third
    for the `z` direction. Missing values will be filled with `1`.

    A different value for the GPU block size and launch bound can be specified for
    maps of dimension 1, 2 or 3 (all maps with higher dimensions are considered
    three dimensional). If no value is specified then the block size `(32, 1, 1)`
    will be used an no launch bound will be be emitted.

    If a Map is found whose iteration size in a particular dimension is smaller than
    the block size, the block size for that dimension is reduced to that value. For
    example, the block size for 2D Maps is `(32, 8, 1)` and a Map that only performs
    four iteration in the second dimension, will get a block size of `(32, 4, 1)`.
    Note that this modification will not influence the launch bound value.

    Args:
        block_size_Xd: The size of a thread block on the GPU for `X` dimensional maps.
        launch_bounds_Xd: The value for the launch bound that should be used for `X`
            dimensional maps.
        launch_factor_Xd: If no `launch_bounds` was given use the number of threads
            in a block multiplied by this number, for maps of dimension `X`.

    Note:
        - You should use the `gt_set_gpu_blocksize()` function.
        - "Over specification" is ignored, i.e. if `(32, 3, 1)` is passed as block
            size for 1 dimensional maps, then it is changed to `(32, 1, 1)`.
        - In case the Map has more than 3 dimension, normal DaCe semantic is used.

    Todo:
        - Turn this into a function.
    """

    _block_size_default: Final[tuple[int, int, int]] = (32, 1, 1)

    block_size_1d = dace_properties.Property(
        dtype=tuple[int, int, int],
        default=_block_size_default,
        setter=_make_gpu_block_parser_for(1),
        getter=_make_gpu_block_getter_for(1),
        desc="Block size for 1 dimensional GPU maps.",
    )
    launch_bounds_1d = dace_properties.Property(
        dtype=str,
        allow_none=True,
        default=None,
        desc="Set the launch bound property for 1 dimensional map.",
    )
    block_size_2d = dace_properties.Property(
        dtype=tuple[int, int, int],
        default=_block_size_default,
        setter=_make_gpu_block_parser_for(2),
        getter=_make_gpu_block_getter_for(2),
        desc="Block size for 2 dimensional GPU maps.",
    )
    launch_bounds_2d = dace_properties.Property(
        dtype=str,
        allow_none=True,
        default=None,
        desc="Set the launch bound property for 2 dimensional map.",
    )
    block_size_3d = dace_properties.Property(
        dtype=tuple[int, int, int],
        default=_block_size_default,
        setter=_make_gpu_block_parser_for(3),
        getter=_make_gpu_block_getter_for(3),
        desc="Block size for 3 dimensional GPU maps.",
    )
    launch_bounds_3d = dace_properties.Property(
        dtype=str,
        allow_none=True,
        default=None,
        desc="Set the launch bound property for 3 dimensional map.",
    )

    # Pattern matching
    map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    def __init__(
        self,
        block_size_1d: Sequence[int | str] | str | None = None,
        block_size_2d: Sequence[int | str] | str | None = None,
        block_size_3d: Sequence[int | str] | str | None = None,
        launch_bounds_1d: int | str | None = None,
        launch_bounds_2d: int | str | None = None,
        launch_bounds_3d: int | str | None = None,
        launch_factor_1d: int | None = None,
        launch_factor_2d: int | None = None,
        launch_factor_3d: int | None = None,
    ) -> None:
        super().__init__()
        if block_size_1d is not None:
            self.block_size_1d = block_size_1d
        if block_size_2d is not None:
            self.block_size_2d = block_size_2d
        if block_size_3d is not None:
            self.block_size_3d = block_size_3d
        self.launch_bounds_1d = _gpu_launch_bound_parser(
            self.block_size_1d, launch_bounds_1d, launch_factor_1d
        )
        self.launch_bounds_2d = _gpu_launch_bound_parser(
            self.block_size_2d, launch_bounds_2d, launch_factor_2d
        )
        self.launch_bounds_3d = _gpu_launch_bound_parser(
            self.block_size_3d, launch_bounds_3d, launch_factor_3d
        )

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.map_entry)]

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        """Test if the block size can be set.

        The function tests:
        - If the block size of the map is already set.
        - If the map is at global scope.
        - If if the schedule of the map is correct.
        """
        scope = graph.scope_dict()
        if scope[self.map_entry] is not None:
            return False
        if self.map_entry.map.schedule not in dace.dtypes.GPU_SCHEDULES:
            return False
        if self.map_entry.map.gpu_block_size is not None:
            return False
        return True

    def apply(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
    ) -> None:
        """Modify the map as requested."""
        gpu_map: dace_nodes.Map = self.map_entry.map
        map_size = gpu_map.range.size()
        num_map_params = len(gpu_map.params)

        # Because of a particularity of the DaCe code generator, the iteration
        #  variable that is associated to the `x` dimension of the block is the
        #  last parameter, i.e. `gpu_map.params[-1]`. The one for `y` the second last.
        if num_map_params == 1:
            block_size = list(self.block_size_1d)
            launch_bounds = self.launch_bounds_1d
            dims_to_inspect = 1
        elif num_map_params == 2:
            block_size = list(self.block_size_2d)
            launch_bounds = self.launch_bounds_2d
            dims_to_inspect = 2
        else:
            block_size = list(self.block_size_3d)
            launch_bounds = self.launch_bounds_3d
            # If there are more than three dimensions DaCe will condense them into
            #  the `z` dimension of the block, so we have to ignore the `z` dimension,
            #  when we modify the block sizes.
            dims_to_inspect = 3 if num_map_params == 3 else 2

        # Cut down the block size.
        # TODO(phimuell): Think if it is useful to also modify the launch bounds.
        # TODO(phimuell): Also think of how to connect this with the loop blocking.
        for i in range(dims_to_inspect):
            map_dim_idx_to_inspect = num_map_params - 1 - i
            if (map_size[map_dim_idx_to_inspect] < block_size[i]) == True:  # noqa: E712 [true-false-comparison]  # SymPy Fancy comparison.
                block_size[i] = map_size[map_dim_idx_to_inspect]

        gpu_map.gpu_block_size = tuple(block_size)
        if launch_bounds is not None:  # Note: empty string has a meaning in DaCe
            gpu_map.gpu_launch_bounds = launch_bounds


def gt_remove_trivial_gpu_maps(
    sdfg: dace.SDFG,
    validate: bool = True,
    validate_all: bool = False,
) -> dace.SDFG:
    """Removes trivial maps that were created by the GPU transformation.

    The main problem is that a Tasklet outside of a Map cannot write into an
    _array_ that is on GPU. `sdfg.apply_gpu_transformations()` will wrap such
    Tasklets in a Map. The `GT4PyMoveTaskletIntoMap` pass, that runs before,
    but only works if the tasklet is adjacent to a map.

    It first tries to promote them such that they can be fused in other non-trivial
    maps, it will then also perform fusion on them, to reduce the number of kernel
    calls.

    Args:
        sdfg: The SDFG that we process.
        validate: Perform validation at the end of the function.
        validate_all: Perform validation also on intermediate steps.

    Todo: Improve this function.
    """

    # First we try to promote and fuse them with other non-trivial maps.
    sdfg.apply_transformations_once_everywhere(
        TrivialGPUMapElimination(
            do_not_fuse=False,
            only_gpu_maps=True,
        ),
        validate=False,
        validate_all=validate_all,
    )
    gtx_transformations.gt_simplify(sdfg, validate=False, validate_all=validate_all)

    # Now we try to fuse them together, however, we restrict the fusion to trivial
    #  GPU map.
    def restrict_to_trivial_gpu_maps(
        self: Union[gtx_transformations.MapFusionVertical, gtx_transformations.MapFusionHorizontal],
        map_node_1: Union[dace_nodes.MapEntry, dace_nodes.MapExit],
        map_entry_2: dace_nodes.MapEntry,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
    ) -> bool:
        map_entry_1 = (
            map_node_1
            if isinstance(map_node_1, dace_nodes.MapEntry)
            else graph.entry_node(map_node_1)
        )
        for map_entry in [map_entry_1, map_entry_2]:
            _map = map_entry.map
            if len(_map.params) != 1:
                return False
            if _map.range[0][0] != _map.range[0][1]:
                return False
            if _map.schedule not in [
                dace.dtypes.ScheduleType.GPU_Device,
                dace.dtypes.ScheduleType.GPU_Default,
            ]:
                return False
        return True

    # TODO(phimuell): Replace this with a more performant loop.
    sdfg.apply_transformations_repeated(
        [
            gtx_transformations.MapFusionVertical(
                only_toplevel_maps=True,
                check_fusion_callback=restrict_to_trivial_gpu_maps,
            ),
            gtx_transformations.MapFusionHorizontal(
                only_toplevel_maps=True,
                check_fusion_callback=restrict_to_trivial_gpu_maps,
            ),
        ],
        validate=False,
        validate_all=validate_all,
    )

    if validate and (not validate_all):
        sdfg.validate()

    return sdfg


@dace_properties.make_properties
class TrivialGPUMapElimination(dace_transformation.SingleStateTransformation):
    """Eliminate certain kind of trivial GPU maps.

    A tasklet outside of map can not write to GPU memory, this can only be done
    from within a map (a scalar is possible). For that reason DaCe's GPU
    transformation wraps such tasklets in trivial maps.
    Under certain condition the transformation will fuse the trivial tasklet with
    a downstream (serial) map.

    Args:
        do_not_fuse: If `True` then the maps are not fused together.
        only_gpu_maps: Only apply to GPU maps; `True` by default.

    Note:
        - This transformation should not be run on its own, instead it
            is run within the context of `gt_gpu_transformation()`.
        - This transformation must be run after the GPU Transformation.

    Todo: Figuring out if this transformation is still needed.
    """

    only_gpu_maps = dace_properties.Property(
        dtype=bool,
        default=True,
        desc="Only promote maps that are GPU maps (debug option).",
    )
    do_not_fuse = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Only perform the promotion, do not fuse.",
    )

    # Pattern Matching
    trivial_map_exit = dace_transformation.PatternNode(dace_nodes.MapExit)
    access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    second_map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    def __init__(
        self,
        do_not_fuse: Optional[bool] = None,
        only_gpu_maps: Optional[bool] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if only_gpu_maps is not None:
            self.only_gpu_maps = only_gpu_maps
        if do_not_fuse is not None:
            self.do_not_fuse = do_not_fuse

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.trivial_map_exit, cls.access_node, cls.second_map_entry
            )
        ]

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        """Tests if the promotion is possible.

        The tests includes:
        - Schedule of the maps.
        - If the map is trivial.
        - Tests if the maps can be fused.
        """
        trivial_map_exit: dace_nodes.MapExit = self.trivial_map_exit
        trivial_map: dace_nodes.Map = trivial_map_exit.map
        trivial_map_entry: dace_nodes.MapEntry = graph.entry_node(trivial_map_exit)
        second_map: dace_nodes.Map = self.second_map_entry.map

        # The kind of maps we are interested only have one parameter.
        if len(trivial_map.params) != 1:
            return False
        for rng in trivial_map.range.ranges:
            if rng[0] != rng[1]:
                return False

        # If we do not not fuse, then the second map can not be trivial.
        #  If we would not prevent that case then we would match these two
        #  maps again and again.
        if self.do_not_fuse and len(second_map.params) <= 1:
            for rng in second_map.range.ranges:
                if rng[0] == rng[1]:
                    return False

        # We now check that the Memlets do not depend on the map parameter.
        #  This is important for the `can_be_applied_to()` check we do below
        #  because we can avoid calling the replace function.
        scope = graph.scope_subgraph(trivial_map_entry)
        trivial_map_param: str = trivial_map.params[0]
        for edge in scope.edges():
            if trivial_map_param in edge.data.free_symbols:
                return False

        # Check if only GPU maps are involved (this is more a testing debug feature).
        if self.only_gpu_maps:
            for map_to_check in [trivial_map, second_map]:
                if map_to_check.schedule not in [
                    dace.dtypes.ScheduleType.GPU_Device,
                    dace.dtypes.ScheduleType.GPU_Default,
                ]:
                    return False

        # Now we check if the two maps can be fused together. For that we have to
        #  do a temporary promotion, it is important that we do not perform the
        #  renaming. If the old symbol is still used, it is used inside a tasklet
        #  so it would show up (temporarily) as free symbol.
        # NOTE: We use the same options as in `MapPromoter`, however, since we do
        #   not have the list of single use data available, we have to specify it.
        org_trivial_map_params = copy.deepcopy(trivial_map.params)
        org_trivial_map_range = copy.deepcopy(trivial_map.range)
        try:
            self._promote_map(graph, replace_trivail_map_parameter=False)
            if not gtx_transformations.MapFusionVertical.can_be_applied_to(
                sdfg=sdfg,
                options={
                    "only_toplevel_maps": True,
                    "require_all_intermediates": True,
                    "require_exclusive_intermediates": True,
                },
                first_map_exit=trivial_map_exit,
                array=self.access_node,
                second_map_entry=self.second_map_entry,
            ):
                return False
        finally:
            trivial_map.params = org_trivial_map_params
            trivial_map.range = org_trivial_map_range

        return True

    def apply(self, graph: Union[dace.SDFGState, dace.SDFG], sdfg: dace.SDFG) -> None:
        """Performs the Map Promoting.

        The function will first perform the promotion of the trivial map and then
        perform the merging of the two maps in one go.
        """
        trivial_map_exit: dace_nodes.MapExit = self.trivial_map_exit
        second_map_entry: dace_nodes.MapEntry = self.second_map_entry
        access_node: dace_nodes.AccessNode = self.access_node

        # Promote the maps.
        self._promote_map(graph)

        # Perform the fusing if requested.
        #  For the selection of the option, see the `can_be_applied()` function.
        if not self.do_not_fuse:
            gtx_transformations.MapFusionVertical.apply_to(
                sdfg=sdfg,
                options={
                    "only_toplevel_maps": True,
                    "require_all_intermediates": True,
                    "require_exclusive_intermediates": True,
                },
                first_map_exit=trivial_map_exit,
                array=access_node,
                second_map_entry=second_map_entry,
                verify=False,  # Do not rerun `can_be_applied()`.
            )

    def _promote_map(
        self,
        state: dace.SDFGState,
        replace_trivail_map_parameter: bool = True,
    ) -> None:
        """Performs the map promoting.

        Essentially this function will copy the parameters and the range from
        the non trivial map (`self.second_map_entry.map`) to the trivial map
        (`self.trivial_map_exit.map`).

        If `replace_trivail_map_parameter` is `True` (the default value), then the
        function will also remove the trivial map parameter with its value.
        """
        assert isinstance(self.trivial_map_exit, dace_nodes.MapExit)
        assert isinstance(self.second_map_entry, dace_nodes.MapEntry)
        assert isinstance(self.access_node, dace_nodes.AccessNode)

        trivial_map_exit: dace_nodes.MapExit = self.trivial_map_exit
        trivial_map: dace_nodes.Map = self.trivial_map_exit.map
        trivial_map_entry: dace_nodes.MapEntry = state.entry_node(trivial_map_exit)
        second_map: dace_nodes.Map = self.second_map_entry.map

        # If requested then replace the map variable with its value.
        if replace_trivail_map_parameter:
            scope = state.scope_subgraph(trivial_map_entry)
            scope.replace(trivial_map.params[0], trivial_map.range[0][0])

        # Now copy parameter and the ranges from the second to the trivial map.
        trivial_map.params = copy.deepcopy(second_map.params)
        trivial_map.range = copy.deepcopy(second_map.range)


def gt_gpu_apply_mempool(sdfg: dace.SDFG) -> None:
    """Enables the stream ordered memory allocator in CUDA code generation.

    It sets the `pool` flag on all transients allocated in GPU global memory.
    The dace code genarator will handle this flag by calling `cudaMallocAsync`
    and `cudaFreeAsync` in the CUDA code for allocation/release of the buffers.

    Args:
        sdfg: The SDFG that should be processed.
    """
    for _, _, desc in sdfg.arrays_recursive():
        if (
            isinstance(desc, dace.data.Array)
            and desc.storage == dace.StorageType.GPU_Global
            and desc.transient
        ):
            desc.pool = True



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
from typing import Any, Optional, cast

from gt4py.eve import NodeTranslator, traits
from gt4py.next import errors
from gt4py.next.ffront import (
    dialect_ast_enums,
    program_ast as past,
    type_specifications as ts_ffront,
)
from gt4py.next.type_system import type_info, type_specifications as ts


def _ensure_no_sliced_field(entry: past.Expr) -> None:
    """
    Check that all arguments are of type past.Name or past.TupleExpr.

    In the latter case, unfold tuple.

    For example, if argument is of type past.Subscript, this function will throw an error as both slicing and domain are being applied
    """
    if not isinstance(entry, past.Name) and not isinstance(entry, past.TupleExpr):
        raise ValueError("Either only domain or slicing allowed.")
    elif isinstance(entry, past.TupleExpr):
        for param in entry.elts:
            _ensure_no_sliced_field(param)


def _is_integral_scalar(expr: past.Expr) -> bool:
    """Check that expression is an integral scalar."""
    return isinstance(expr.type, ts.ScalarType) and type_info.is_integral(expr.type)


def _validate_operator_call(new_func: past.Name, new_kwargs: dict) -> None:
    """
    Perform checks for domain and output field types.

    Keyword `out` has to be present in function call.

    Domain has to be of type dictionary, including dimensions with values expressed as tuples of 2 numbers.
    """
    if not isinstance(new_func.type, (ts_ffront.FieldOperatorType, ts_ffront.ScanOperatorType)):
        raise ValueError(
            f"Only calls to 'FieldOperators' and 'ScanOperators' "
            f"allowed in 'Program', got '{new_func.type}'."
        )

    if "out" not in new_kwargs:
        raise ValueError("Missing required keyword argument 'out'.")
    if "domain" in new_kwargs:
        _ensure_no_sliced_field(new_kwargs["out"])

        domain_kwarg = new_kwargs["domain"]
        if not isinstance(domain_kwarg, past.Dict):
            raise ValueError(f"Only Dictionaries allowed in 'domain', got '{type(domain_kwarg)}'.")

        if len(domain_kwarg.values_) == 0 and len(domain_kwarg.keys_) == 0:
            raise ValueError("Empty domain not allowed.")

        for dim in domain_kwarg.keys_:
            if not isinstance(dim.type, ts.DimensionType):
                raise ValueError(
                    f"Only 'Dimension' allowed in domain dictionary keys, got '{dim}' which is of type '{dim.type}'."
                )
        for domain_values in domain_kwarg.values_:
            if len(domain_values.elts) != 2:
                raise ValueError(
                    f"Only 2 values allowed in domain range, got {len(domain_values.elts)}."
                )
            if not _is_integral_scalar(domain_values.elts[0]) or not _is_integral_scalar(
                domain_values.elts[1]
            ):
                raise ValueError(
                    f"Only integer values allowed in domain range, got '{domain_values.elts[0].type}' and '{domain_values.elts[1].type}'."
                )


class ProgramTypeDeduction(traits.VisitorWithSymbolTableTrait, NodeTranslator):
    @classmethod
    def apply(cls, node: past.Program) -> past.Program:
        return cls().visit(node)

    def visit_Program(self, node: past.Program, **kwargs: Any) -> past.Program:
        params = self.visit(node.params, **kwargs)

        definition_type = ts.FunctionType(
            pos_only_args=[],
            pos_or_kw_args={str(param.id): param.type for param in params},
            kw_only_args={},
            returns=ts.VoidType(),
        )
        return past.Program(
            id=self.visit(node.id, **kwargs),
            type=ts_ffront.ProgramType(definition=definition_type),
            params=params,
            body=self.visit(node.body, **kwargs),
            closure_vars=self.visit(node.closure_vars, **kwargs),
            location=node.location,
        )

    def visit_Slice(self, node: past.Slice, **kwargs: Any) -> past.Slice:
        return past.Slice(
            lower=self.visit(node.lower, **kwargs),
            upper=self.visit(node.upper, **kwargs),
            step=self.visit(node.step, **kwargs),
            type=ts.DeferredType(constraint=None),
            location=node.location,
        )

    def visit_Subscript(self, node: past.Subscript, **kwargs: Any) -> past.Subscript:
        value = self.visit(node.value, **kwargs)
        return past.Subscript(
            value=value,
            slice_=self.visit(node.slice_, **kwargs),
            type=value.type,
            location=node.location,
        )

    def visit_Attribute(self, node: past.Attribute, **kwargs: Any) -> past.Attribute:
        new_value = self.visit(node.value, **kwargs)
        return past.Attribute(
            value=new_value,
            attr=node.attr,
            location=node.location,
            type=getattr(new_value.type, node.attr),
        )

    def visit_TupleExpr(self, node: past.TupleExpr, **kwargs: Any) -> past.TupleExpr:
        elts = self.visit(node.elts, **kwargs)
        return past.TupleExpr(
            elts=elts, type=ts.TupleType(types=[el.type for el in elts]), location=node.location
        )

    def _deduce_binop_type(
        self, node: past.BinOp, *, left: past.Expr, right: past.Expr, **kwargs: Any
    ) -> Optional[ts.TypeSpec]:
        logical_ops = {
            dialect_ast_enums.BinaryOperator.BIT_AND,
            dialect_ast_enums.BinaryOperator.BIT_OR,
        }
        is_compatible = type_info.is_logical if node.op in logical_ops else type_info.is_arithmetic

        # check both types compatible
        for arg in (left, right):
            if not isinstance(arg.type, ts.ScalarType) or not is_compatible(arg.type):
                raise errors.DSLError(
                    arg.location, f"Type '{arg.type}' can not be used in operator '{node.op}'."
                )

        left_type = cast(ts.ScalarType, left.type)
        right_type = cast(ts.ScalarType, right.type)

        if node.op == dialect_ast_enums.BinaryOperator.POW:
            return left_type

        if node.op == dialect_ast_enums.BinaryOperator.MOD and not type_info.is_integral(
            right_type
        ):
            raise errors.DSLError(
                arg.location,
                f"Type '{right_type}' can not be used in operator '{node.op}', it only accepts 'int'.",
            )

        try:
            return type_info.promote(left_type, right_type)
        except ValueError as ex:
            raise errors.DSLError(
                node.location,
                f"Could not promote '{left_type}' and '{right_type}' to common type"
                f" in call to '{node.op}'.",
            ) from ex

    def visit_BinOp(self, node: past.BinOp, **kwargs: Any) -> past.BinOp:
        new_left = self.visit(node.left, **kwargs)
        new_right = self.visit(node.right, **kwargs)
        new_type = self._deduce_binop_type(node, left=new_left, right=new_right)
        return past.BinOp(
            op=node.op, left=new_left, right=new_right, location=node.location, type=new_type
        )

    def visit_Call(self, node: past.Call, **kwargs: Any) -> past.Call:
        new_func = self.visit(node.func, **kwargs)
        new_args = self.visit(node.args, **kwargs)
        new_kwargs = self.visit(node.kwargs, **kwargs)

        try:
            is_operator = isinstance(
                new_func.type, (ts_ffront.FieldOperatorType, ts_ffront.ScanOperatorType)
            )
            if is_operator:
                _validate_operator_call(new_func, new_kwargs)
            arg_types = [arg.type for arg in new_args]
            kwarg_types = {
                name: expr.type
                for name, expr in new_kwargs.items()
                if not (is_operator and name in ["out", "domain"])
            }

            type_info.accepts_args(
                new_func.type, with_args=arg_types, with_kwargs=kwarg_types, raise_exception=True
            )
            return_type = ts.VoidType()
            if is_operator:
                operator_return_type = type_info.return_type(
                    new_func.type, with_args=arg_types, with_kwargs=kwarg_types
                )
                if operator_return_type != new_kwargs["out"].type:
                    raise ValueError(
                        "Expected keyword argument 'out' to be of "
                        f"type '{operator_return_type}', got "
                        f"'{new_kwargs['out'].type}'."
                    )
            elif new_func.id in ["minimum", "maximum"]:
                if arg_types[0] != arg_types[1]:
                    raise ValueError(
                        f"First and second argument in '{new_func.id}' must be of the same type."
                        f"Got '{arg_types[0]}' and '{arg_types[1]}'."
                    )
                return_type = arg_types[0]
            else:
                raise AssertionError(
                    "Only calls to 'FieldOperator', 'ScanOperator' or 'minimum' and 'maximum' builtins allowed."
                )

        except ValueError as ex:
            raise errors.DSLError(node.location, f"Invalid call to '{node.func.id}'.\n{ex}") from ex

        return past.Call(
            func=new_func,
            args=new_args,
            kwargs=new_kwargs,
            type=return_type,
            location=node.location,
        )

    def visit_Name(self, node: past.Name, **kwargs: Any) -> past.Name:
        symtable = kwargs["symtable"]
        if node.id not in symtable or symtable[node.id].type is None:
            raise errors.DSLError(node.location, f"Undeclared or untyped symbol '{node.id}'.")

        return past.Name(id=node.id, type=symtable[node.id].type, location=node.location)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause


import copy

import dace
from dace import (
    data as dace_data,
    dtypes as dace_dtypes,
    symbolic as dace_symbolic,
    transformation as dace_transformation,
)
from dace.sdfg import graph as dace_graph, nodes as dace_nodes

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


def gt_create_local_double_buffering(
    sdfg: dace.SDFG,
) -> int:
    """Modifies the SDFG such that point wise data dependencies are stable.

    Rule 3 of the ADR18, guarantees that if data is input and output to a map,
    then it must be a non transient array and it must only have point wise
    dependency. This means that every index that is read is also written by
    the same thread and no other thread reads or writes to the same location.
    However, because the dataflow inside a map is partially asynchronous
    it might happen if something is read multiple times, i.e. Tasklets,
    the data might already be overwritten.
    This function will scan the SDFG for potential cases and insert an
    access node to cache this read. This is essentially a double buffer, but
    it is not needed that the whole data is stored, but only the working set
    of a single thread.

    Todo:
        Simplify this function.
    """
    processed_maps = 0
    for nsdfg in sdfg.all_sdfgs_recursive():
        processed_maps += _create_local_double_buffering_non_recursive(nsdfg)
    return processed_maps


def _create_local_double_buffering_non_recursive(
    sdfg: dace.SDFG,
) -> int:
    """Implementation of the point wise transformation.

    This function does not handle nested SDFGs.
    """
    # First we call `EdgeConsolidation`, because of that we know that
    #  every incoming edge of a `MapEntry` refers to distinct data.
    #  We do this to simplify our implementation.
    edge_consolidation = dace_transformation.passes.ConsolidateEdges()
    edge_consolidation.apply_pass(sdfg, None)

    processed_maps = 0
    for state in sdfg.states():
        assert isinstance(state, dace.SDFGState)
        scope_dict = state.scope_dict()
        for node in state.nodes():
            if not isinstance(node, dace_nodes.MapEntry):
                continue
            if scope_dict[node] is not None:
                continue
            inout_nodes = _check_if_map_must_be_handled(
                map_entry=node,
                state=state,
                sdfg=sdfg,
            )
            if inout_nodes is not None:
                processed_maps += _add_local_double_buffering_to(
                    map_entry=node,
                    inout_nodes=inout_nodes,
                    state=state,
                    sdfg=sdfg,
                )
    return processed_maps


def _add_local_double_buffering_to(
    inout_nodes: dict[str, tuple[dace_nodes.AccessNode, dace_nodes.AccessNode]],
    map_entry: dace_nodes.MapEntry,
    state: dace.SDFGState,
    sdfg: dace.SDFG,
) -> int:
    """Adds the double buffering to `map_entry` for `inout_nodes`.

    The function assumes that there is only in incoming edge per data
    descriptor at the map entry. If the data is needed multiple times,
    then the distribution must be done inside the map.

    The function will now channel all reads to the data descriptor
    through an access node, this ensures that the read happens
    before the write.
    """
    processed_maps = 0
    for inout_node in inout_nodes.values():
        _add_local_double_buffering_to_single_data(
            map_entry=map_entry,
            inout_node=inout_node,
            state=state,
            sdfg=sdfg,
        )
        processed_maps += 1
    return processed_maps


def _add_local_double_buffering_to_single_data(
    inout_node: tuple[dace_nodes.AccessNode, dace_nodes.AccessNode],
    map_entry: dace_nodes.MapEntry,
    state: dace.SDFGState,
    sdfg: dace.SDFG,
) -> None:
    """Adds the local double buffering for a single data."""
    map_exit: dace_nodes.MapExit = state.exit_node(map_entry)
    input_node, output_node = inout_node
    input_edges = state.edges_between(input_node, map_entry)
    output_edges = state.edges_between(map_exit, output_node)
    assert len(input_edges) == 1
    assert len(output_edges) == 1
    inner_read_edges = _get_inner_edges(input_edges[0], map_entry, state, False)
    inner_write_edges = _get_inner_edges(output_edges[0], map_exit, state, True)

    # Create the data container, we know that it is a scalar from the filtering
    assert _is_inner_read_a_scalar(inner_read_edges[0], state, sdfg)
    new_double_inner_buff_name: str = f"__inner_double_buffer_for_{input_node.data}"
    new_double_inner_buff_name, _ = sdfg.add_scalar(
        new_double_inner_buff_name,
        dtype=input_node.desc(sdfg).dtype,
        transient=True,
        storage=dace_dtypes.StorageType.Register,
        find_new_name=True,
    )
    new_double_inner_buff_node = state.add_access(
        new_double_inner_buff_name, copy.copy(input_node.debuginfo)
    )

    # Now reroute the data flow through the new access node.
    reconfigured_dataflow: set[tuple[dace_nodes.Node, str]] = set()
    for old_inner_read_edge in inner_read_edges:
        # TODO(phimuell): Handle the case the memlet is "fancy"
        new_edge = state.add_edge(
            new_double_inner_buff_node,
            None,
            old_inner_read_edge.dst,
            old_inner_read_edge.dst_conn,
            dace.Memlet(
                data=new_double_inner_buff_name,
                subset="0",
                other_subset=copy.deepcopy(
                    old_inner_read_edge.data.get_dst_subset(old_inner_read_edge, state)
                ),
            ),
        )
        state.remove_edge(old_inner_read_edge)
        if (new_edge.dst, new_edge.dst_conn) not in reconfigured_dataflow:
            gtx_transformations.utils.reconfigure_dataflow_after_rerouting(
                is_producer_edge=False,
                new_edge=new_edge,
                ss_offset=None,
                state=state,
                sdfg=sdfg,
                old_node=input_node,
                new_node=new_double_inner_buff_node,
            )
            reconfigured_dataflow.add((new_edge.dst, new_edge.dst_conn))

    # Now create a connection between the map entry and the intermediate node.
    state.add_edge(
        map_entry,
        inner_read_edges[0].src_conn,
        new_double_inner_buff_node,
        None,
        dace.Memlet(
            data=input_node.data,
            subset=copy.deepcopy(
                inner_read_edges[0].data.get_src_subset(inner_read_edges[0], state)
            ),
            other_subset="0",
        ),
    )

    # To really ensure that a read happens before a write, we have to sequence
    #  the read first. We do this by connecting the double buffer node with
    #  empty Memlets to the last row of nodes that writes to the global buffer.
    #  This is needed to handle the case that some other data path performs the
    #  write.
    # TODO(phimuell): Add a test that only performs this when it is really needed.
    for inner_write_edge in inner_write_edges:
        state.add_nedge(
            new_double_inner_buff_node,
            inner_write_edge.src,
            dace.Memlet(),
        )


def _is_inner_read_a_scalar(
    inner_reading_edge: dace_graph.MultiConnectorEdge,
    state: dace.SDFGState,
    sdfg: dace.SDFG,
) -> bool:
    """Tests if the read by `inner_reading_edge` is a scalar.

    A condition for being pointwise is that only one scalar is read, so this function
    checks this.

    Args:
        inner_reading_edge: The edge at the inside of the Map that reads from data
            that is connected to the MapEntry from the outside.
        state: The state in which we operate.
        sdfg: The SDFG we operate on.
    """
    new_double_inner_buff_shape_raw = dace_symbolic.overapproximate(
        inner_reading_edge.data.get_src_subset(inner_reading_edge, state).size()
    )

    # We can decay it into a scalar if all shapes have size 1.
    for proposed_dim_size in new_double_inner_buff_shape_raw:
        if (proposed_dim_size == 1) == False:  # noqa: E712 [true-false-comparison]  # SymPy Fancy comparison.
            return False
    return True


def _check_if_map_must_be_handled_classify_adjacent_access_node(
    data_node: dace_nodes.AccessNode,
    sdfg: dace.SDFG,
    known_nodes: dict[str, dace_nodes.AccessNode],
) -> bool:
    """Internal function used by `_check_if_map_must_be_handled()` to classify nodes.

    If the function returns `True` it means that the input/output, does not
    violates an internal constraint, i.e. can be handled by
    `_ensure_that_map_is_pointwise()`. If appropriate the function will add the
    node to `known_nodes`. I.e. in case of a transient the function will return
    `True` but will not add it to `known_nodes`.
    """

    # This case is indicating that the `ConsolidateEdges` has not fully worked.
    #  Currently the transformation implementation assumes that this is the
    #  case, so we can not handle this case.
    # TODO(phimuell): Implement this case.
    if data_node.data in known_nodes:
        return False
    data_desc: dace_data.Data = data_node.desc(sdfg)

    # The conflict can only occur for global data, because transients
    #  are only written once.
    if data_desc.transient:
        return False

    # Currently we do not handle view, as they need to be traced.
    #  TODO(phimuell): Implement
    if gtx_transformations.utils.is_view(data_desc, sdfg):
        return False

    # TODO(phimuell): Check if there is a access node on the inner side, then we do not have to do it.

    # Now add the node to the list.
    assert all(data_node is not known_node for known_node in known_nodes.values())
    known_nodes[data_node.data] = data_node
    return True


def _get_inner_edges(
    outer_edge: dace.sdfg.graph.MultiConnectorEdge,
    scope_node: dace_nodes.MapExit | dace_nodes.MapEntry,
    state: dace.SDFG,
    outgoing_edge: bool,
) -> list[dace.sdfg.graph.MultiConnectorEdge]:
    """Gets the edges on the inside of a map."""
    if outgoing_edge:
        assert isinstance(scope_node, dace_nodes.MapExit)
        conn_name = outer_edge.src_conn[4:]
        return list(state.in_edges_by_connector(scope_node, connector="IN_" + conn_name))
    else:
        assert isinstance(scope_node, dace_nodes.MapEntry)
        conn_name = outer_edge.dst_conn[3:]
        return list(state.out_edges_by_connector(scope_node, connector="OUT_" + conn_name))


def _check_if_map_must_be_handled(
    map_entry: dace_nodes.MapEntry,
    state: dace.SDFGState,
    sdfg: dace.SDFG,
) -> None | dict[str, tuple[dace_nodes.AccessNode, dace_nodes.AccessNode]]:
    """Check if the map should be processed to uphold rule 3.

    Essentially the function will check if there is a potential read-write
    conflict. The function assumes that `ConsolidateEdges` has already run.

    If there is a possible data race the function will return a `dict`, that
    maps the name of the data to the access nodes that are used as input and
    output to the Map.

    Otherwise, the function returns `None`. It is, however, important that
    `None` does not means that there is no possible race condition. It could
    also means that the function that implements the buffering, i.e.
    `_ensure_that_map_is_pointwise()`, is unable to handle this case.

    Todo:
        Improve the function
    """
    map_exit: dace_nodes.MapExit = state.exit_node(map_entry)

    # Find all the data that is accessed. Views are resolved.
    input_datas: dict[str, dace_nodes.AccessNode] = {}
    output_datas: dict[str, dace_nodes.AccessNode] = {}

    # Determine which nodes are possible conflicting.
    for in_edge in state.in_edges(map_entry):
        if in_edge.data.is_empty():
            continue
        if not isinstance(in_edge.src, dace_nodes.AccessNode):
            # TODO(phiumuell): Figuring out what this case means
            continue
        if in_edge.dst_conn and not in_edge.dst_conn.startswith("IN_"):
            # TODO(phimuell): It is very unlikely that a Dynamic Map Range causes
            #   this particular data race, so we ignore it for the time being.
            continue
        if not _check_if_map_must_be_handled_classify_adjacent_access_node(
            data_node=in_edge.src,
            sdfg=sdfg,
            known_nodes=input_datas,
        ):
            continue
    for out_edge in state.out_edges(map_exit):
        if out_edge.data.is_empty():
            continue
        if not isinstance(out_edge.dst, dace_nodes.AccessNode):
            # TODO(phiumuell): Figuring out what this case means
            continue
        if not _check_if_map_must_be_handled_classify_adjacent_access_node(
            data_node=out_edge.dst,
            sdfg=sdfg,
            known_nodes=output_datas,
        ):
            continue

    # Double buffering is only needed if there inout arguments.
    inout_datas: dict[str, tuple[dace_nodes.AccessNode, dace_nodes.AccessNode]] = {
        dname: (input_datas[dname], output_datas[dname])
        for dname in input_datas
        if dname in output_datas
    }
    if len(inout_datas) == 0:
        return None

    # TODO(phimuell): What about the case that some data descriptor needs double
    #   buffering, but some do not?
    for inout_data_name in list(inout_datas.keys()):
        input_node, output_node = inout_datas[inout_data_name]
        input_edges = state.edges_between(input_node, map_entry)
        output_edges = state.edges_between(map_exit, output_node)
        assert len(input_edges) == 1, (
            f"Expected a single connection between input node and map entry, but found {len(input_edges)}."
        )
        assert len(output_edges) == 1, (
            f"Expected a single connection between map exit and write back node, but found {len(output_edges)}."
        )

        # If there is only one edge on the inside of the map, that goes into an
        #  AccessNode, then we assume it is double buffered.
        inner_read_edges = _get_inner_edges(input_edges[0], map_entry, state, False)
        if (
            len(inner_read_edges) == 1
            and isinstance(inner_read_edges[0].dst, dace_nodes.AccessNode)
            and not gtx_transformations.utils.is_view(inner_read_edges[0].dst, sdfg)
        ):
            inout_datas.pop(inout_data_name)
            continue

        # For being pointwise the read must only be a scalar, so by definition
        #  if this does not hold, we can ignore it.
        # NOTE: This is not always true, in case of nested Maps, for example the
        #   ones that are generated by `LoopBlocking`, but this transformation
        #   should run before that anyway and it is invalid to run auto optimizer
        #   multiple time on the SDFG.
        if not all(
            _is_inner_read_a_scalar(inner_read_edge, state, sdfg)
            for inner_read_edge in inner_read_edges
        ):
            inout_datas.pop(inout_data_name)
            continue

        inner_read_subsets = [
            inner_read_edge.data.get_src_subset(inner_read_edge, state)
            for inner_read_edge in inner_read_edges
        ]
        assert all(inner_read_subset is not None for inner_read_subset in inner_read_subsets)
        inner_write_subsets = [
            inner_write_edge.data.get_dst_subset(inner_write_edge, state)
            for inner_write_edge in _get_inner_edges(output_edges[0], map_exit, state, True)
        ]
        # TODO(phimuell): Also implement a check that the volume equals the size of the subset.
        assert all(inner_write_subset is not None for inner_write_subset in inner_write_subsets)

        # For being point wise the subsets must be compatible. The correct check would be:
        #   - The write sets are unique.
        #   - For every read subset there exists one matching write subset. It could
        #       be that there are many equivalent read subsets.
        #   - For every write subset there exists at least one matching read subset.
        #  The current implementation only checks if all are the same.
        # TODO(phimuell): Implement the real check.
        all_inner_subsets = inner_read_subsets + inner_write_subsets
        if not all(
            all_inner_subsets[0] == all_inner_subsets[i] for i in range(1, len(all_inner_subsets))
        ):
            return None

    if len(inout_datas) == 0:
        return None

    return inout_datas



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import contextvars
import dataclasses
import json
import pathlib
import sys
from collections.abc import Generator, MutableSequence, Sequence

import numpy as np

from gt4py.eve import utils
from gt4py.eve.extended_typing import Final, TypeVar
from gt4py.next import config


# Common metric names
COMPUTE_METRIC: Final = sys.intern("compute")
TOTAL_METRIC: Final = sys.intern("total")


# Metric collection levels
DISABLED: Final = 0
MINIMAL: Final = 1
PERFORMANCE: Final = 10
INFO: Final = 30
VERBOSE: Final = 50
ALL: Final = 100


@dataclasses.dataclass(frozen=True)
class Metric:
    """
    A class to collect and analyze samples of a metric.

    Examples:
        >>> metric = Metric(name="execution_time")
        >>> metric.add_sample(0.1)
        >>> metric.add_sample(0.2)
        >>> print(f"{metric.mean:.2}")
        0.15
        >>> print(metric)
        1.50000e-01 +/- 7.07107e-02
    """

    name: str | None = None
    samples: MutableSequence[float] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        if self.name:
            object.__setattr__(self, "name", sys.intern(self.name))

    @property
    def mean(self) -> np.floating:
        if len(self.samples) == 0:
            raise ValueError("Cannot compute mean of empty sample list.")
        return np.mean(self.samples)

    @property
    def std(self) -> np.floating:
        if len(self.samples) == 0:
            raise ValueError("Cannot compute std of empty sample list.")
        return np.std(self.samples, ddof=1)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, values={self.samples})"

    def __str__(self) -> str:
        return f"{self.mean:.5e} +/- {self.std:.5e}"

    def add_sample(self, sample: float) -> None:
        self.samples.append(sample)


class MetricCollection(utils.CustomDefaultDictBase[str, Metric]):
    """
    A collection of metrics, organized as a mapping from metric names to `Metric` objects.

    Empty `Metric` instances are created automatically when accessing keys
    that do not exist.

    Example:
        >>> metrics = MetricCollection()
        >>> metrics["execution_time"].add_sample(0.1)
        >>> metrics["execution_time"].add_sample(0.2)
        >>> metrics["execution_time"].samples
        [0.1, 0.2]
    """

    def value_factory(self, key: str) -> Metric:
        return Metric(name=key)

    def add_sample(self, name: str, sample: float) -> None:
        self[name].samples.append(sample)


_KeyT = TypeVar("_KeyT", default=str)


class MetricCollectionStore(utils.CustomDefaultDictBase[_KeyT, MetricCollection]):
    """
    A dictionary-like store for metric collections.

    Empty `MetricCollection` instances are created automatically when accessing keys
    that don't exist yet.

    Example:
        >>> store = MetricCollectionStore()
        >>> store["program1"]["execution_time"].add_sample(0.1)
        >>> store["program1"]["execution_time"].samples
        [0.1]

        >>> store["program2"]["execution_time"].add_sample(0.2)
        >>> store["program2"]["foo_time"].add_sample(2.3)
        >>> store.metric_names
        ['execution_time', 'foo_time']
    """

    def value_factory(self, key: _KeyT) -> MetricCollection:
        return MetricCollection()

    @property
    def metric_names(self) -> Sequence[str]:
        """Returns a list of all metric names across all collections in the store."""

        return list(
            dict.fromkeys(
                (name for collection in self.values() for name in collection.keys())
            ).keys()
        )


# Context variable storing the active metric collection of the current program.
_active_metric_collection: contextvars.ContextVar[MetricCollection | None] = contextvars.ContextVar(
    "active_metric_collection", default=None
)

#: Global metric collection store for the entire program.
program_metrics: MetricCollectionStore[str] = MetricCollectionStore()


def get_active_metric_collection() -> MetricCollection | None:
    """Retrieve the active metric collection for the current program."""
    return _active_metric_collection.get()


@contextlib.contextmanager
def metrics_collection(key: str) -> Generator[MetricCollection | None, None, None]:
    current_metrics: MetricCollection | None = None
    reset_token: contextvars.Token | None = None
    if config.COLLECT_METRICS_LEVEL:
        current_metrics = program_metrics[key]
        assert _active_metric_collection.get() in [None, current_metrics]
        reset_token = _active_metric_collection.set(current_metrics)

    try:
        yield current_metrics

    finally:
        if reset_token is not None:
            _active_metric_collection.reset(reset_token)


def dumps(metric_cs: MetricCollectionStore | None = None) -> str:
    """
    Format the metrics in the collection store as a string table.

    This function generates a formatted string representation of the metrics
    in the collection store. Each row represents a program, and each column
    represents a metric, showing both the mean value and standard deviation.

    If no explicit collection store is provided, uses the global `program_metrics`.
    """
    if metric_cs is None:
        metric_cs = program_metrics

    title_program = "program"
    title_cols = max(len(k) for k in (*metric_cs.keys(), title_program)) + 1
    metric_names = metric_cs.metric_names
    width = max(len(name) for name in (*metric_names, ""))
    titles = (f"{name:<{width}}  {'+/-':<{width}}" for name in metric_names)
    header = f"{title_program:{title_cols}} {'  '.join(titles)}"

    rows = []
    for program in metric_cs.keys():
        cells = []
        for metric_name in metric_names:
            if metric_name in metric_cs[program]:
                cells.append(
                    f"{metric_cs[program][metric_name].mean:<{width}.5e}  "
                    f"{metric_cs[program][metric_name].std:<{width}.5e}"
                )
            else:
                cells.append(f"{'N/A':<{width}} {'N/A':<{width}}")
        rows.append(f"{program:{title_cols}} {'  '.join(cells)}")

    return "\n".join(["", header, *rows, ""])


def dump(filename: str | pathlib.Path, metric_cs: MetricCollectionStore | None = None) -> None:
    pathlib.Path(filename).write_text(dumps(metric_cs))


def dumps_json(metric_cs: MetricCollectionStore | None = None) -> str:
    """
    Export metrics as a JSON string with structure: {program: {metric_name: [samples]}}

    If no explicit collection store is provided, uses the global `program_metrics`.
    """
    if metric_cs is None:
        metric_cs = program_metrics

    return json.dumps(
        {
            program: {name: metric.samples for name, metric in collection.items()}
            for program, collection in metric_cs.items()
        },
    )


def dump_json(filename: str | pathlib.Path, metric_cs: MetricCollectionStore | None = None) -> None:
    pathlib.Path(filename).write_text(dumps_json(metric_cs))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses

from gt4py.eve import NodeTranslator, traits
from gt4py.eve.utils import UIDGenerator
from gt4py.next.iterator import ir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm
from gt4py.next.iterator.transforms import inline_lambdas


@dataclasses.dataclass(frozen=True)
class FuseMaps(traits.PreserveLocationVisitor, traits.VisitorWithSymbolTableTrait, NodeTranslator):
    """
    Fuses nested `map_`s.

    Preconditions:
      - `FunctionDefinitions` are inlined
      - Pass must be only constructed once (and `_fuse_mapsX` are reserved symbols)

    Example:
        map(λ(x, y)->f(x, y))(a, map(λ(z, w)->g(z, w))(b, c))
    to
        map(λ(a, b, c) → f(a, g(b, c)))(a, b, c)

        reduce(λ(x, y) → f(x, y), init)(map_(g(z, w))(a, b))
    to
        reduce(λ(x, y, z) → f(x, g(y, z)), init)(a, b)
    """

    uids: UIDGenerator = dataclasses.field(init=False, repr=False, default_factory=UIDGenerator)

    def _as_lambda(self, fun: ir.SymRef | ir.Lambda, param_count: int) -> ir.Lambda:
        # if fun is already a Lambda we still wrap it to get unique symbol names to avoid symbol clashes
        params = [
            ir.Sym(id=self.uids.sequential_id(prefix="_fuse_maps")) for _ in range(param_count)
        ]
        return ir.Lambda(
            params=params,
            expr=ir.FunCall(fun=fun, args=[ir.SymRef(id=p.id) for p in params]),
            location=fun.location,
        )

    def visit_FunCall(self, node: ir.FunCall, **kwargs):
        node = self.generic_visit(node)
        if cpm.is_applied_map(node) or cpm.is_applied_reduce(node):
            if any(cpm.is_applied_map(arg) for arg in node.args):
                first_param = (
                    0 if cpm.is_applied_map(node) else 1
                )  # index of the first param of op that maps to args (0 for map, 1 for reduce)
                assert isinstance(node.fun, ir.FunCall)
                assert isinstance(node.fun.args[0], (ir.Lambda, ir.SymRef))
                outer_op = self._as_lambda(node.fun.args[0], len(node.args) + first_param)

                inlined_args = []
                new_params = []
                new_args = []
                if cpm.is_applied_reduce(node):
                    # param corresponding to reduce acc
                    inlined_args.append(ir.SymRef(id=outer_op.params[0].id))
                    new_params.append(outer_op.params[0])

                for i in range(len(node.args)):
                    if cpm.is_applied_map(node.args[i]):
                        map_call = node.args[i]
                        assert isinstance(map_call, ir.FunCall)
                        assert isinstance(map_call.fun, ir.FunCall)
                        assert isinstance(map_call.fun.args[0], (ir.Lambda, ir.SymRef))
                        inner_op = self._as_lambda(map_call.fun.args[0], len(map_call.args))
                        inlined_args.append(
                            inline_lambdas.inline_lambda(
                                ir.FunCall(
                                    fun=inner_op,
                                    args=[ir.SymRef(id=param.id) for param in inner_op.params],
                                    location=node.location,
                                )
                            )
                        )
                        new_params.extend(inner_op.params)
                        new_args.extend(map_call.args)
                    else:
                        inlined_args.append(ir.SymRef(id=outer_op.params[i + first_param].id))
                        new_params.append(outer_op.params[i + first_param])
                        new_args.append(node.args[i])
                new_body = ir.FunCall(fun=outer_op, args=inlined_args)
                new_body = inline_lambdas.inline_lambda(
                    new_body
                )  # removes one level of nesting (the recursive inliner could simplify more, however this can also be done on the full tree later)
                new_op = ir.Lambda(params=new_params, expr=new_body)
                if cpm.is_applied_map(node):
                    return ir.FunCall(
                        fun=ir.FunCall(fun=ir.SymRef(id="map_"), args=[new_op]), args=new_args
                    )
                else:  # is_applied_reduce(node)
                    return ir.FunCall(
                        fun=ir.FunCall(fun=ir.SymRef(id="reduce"), args=[new_op, node.fun.args[1]]),
                        args=new_args,
                    )
        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import typing
from typing import Generic

from gt4py.next.otf import workflow


PrgT = typing.TypeVar("PrgT")
ArgT = typing.TypeVar("ArgT")
StartT = typing.TypeVar("StartT")
EndT = typing.TypeVar("EndT")


@dataclasses.dataclass
class CompilableProgram(Generic[PrgT, ArgT]):
    data: PrgT
    args: ArgT


@dataclasses.dataclass(frozen=True)
class DataOnlyAdapter(
    workflow.ChainableWorkflowMixin,
    workflow.ReplaceEnabledWorkflowMixin,
    workflow.Workflow[CompilableProgram[StartT, ArgT], CompilableProgram[EndT, ArgT]],
    Generic[ArgT, StartT, EndT],
):
    step: workflow.Workflow[StartT, EndT]

    def __call__(self, inp: CompilableProgram[StartT, ArgT]) -> CompilableProgram[EndT, ArgT]:
        return CompilableProgram(data=self.step(inp.data), args=inp.args)


@dataclasses.dataclass(frozen=True)
class ArgsOnlyAdapter(
    workflow.ChainableWorkflowMixin,
    workflow.ReplaceEnabledWorkflowMixin,
    workflow.Workflow[CompilableProgram[PrgT, StartT], CompilableProgram[PrgT, EndT]],
    Generic[PrgT, StartT, EndT],
):
    step: workflow.Workflow[StartT, EndT]

    def __call__(self, inp: CompilableProgram[PrgT, StartT]) -> CompilableProgram[PrgT, EndT]:
        return CompilableProgram(data=inp.data, args=self.step(inp.args))


@dataclasses.dataclass(frozen=True)
class StripArgsAdapter(
    workflow.ChainableWorkflowMixin,
    workflow.ReplaceEnabledWorkflowMixin,
    workflow.Workflow[CompilableProgram[StartT, ArgT], EndT],
    Generic[ArgT, StartT, EndT],
):
    step: workflow.Workflow[StartT, EndT]

    def __call__(self, inp: CompilableProgram[StartT, ArgT]) -> EndT:
        return self.step(inp.data)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import importlib
import importlib.util
import pathlib
from types import ModuleType


def import_from_path(module_file: pathlib.Path) -> ModuleType:
    """Import all function objects from a Python module and return a mapping {function_name: object}."""
    module_name = module_file.name.split(".")[0]

    error_msg = f"Could not load module named {module_name} from {module_file}"
    spec = importlib.util.spec_from_file_location(module_name, module_file)
    if not spec or not spec.loader:
        raise ModuleNotFoundError(error_msg)
    try:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except ImportError as ex:
        raise ModuleNotFoundError(error_msg) from ex

    return module



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Caching for compiled backend artifacts."""

import hashlib
import pathlib
import tempfile

from gt4py.next import config
from gt4py.next.otf import stages
from gt4py.next.otf.binding import interface


_session_cache_dir = tempfile.TemporaryDirectory(prefix="gt4py_session_")

_session_cache_dir_path = pathlib.Path(_session_cache_dir.name)


def _serialize_param(parameter: interface.Parameter) -> str:
    return f"{parameter.name}: {parameter.type_!s}"


def _serialize_library_dependency(dependency: interface.LibraryDependency) -> str:
    return f"{dependency.name}/{dependency.version}"


def _serialize_source(source: stages.ProgramSource) -> str:
    parameters = [_serialize_param(param) for param in source.entry_point.parameters]
    dependencies = [_serialize_library_dependency(dep) for dep in source.library_deps]
    return f"""\
    language: {source.language}
    name: {source.entry_point.name}
    params: {", ".join(parameters)}
    deps: {", ".join(dependencies)}
    src: {source.source_code}
    implicit_domain: {source.implicit_domain}
    """


def _cache_folder_name(source: stages.ProgramSource) -> str:
    serialized = _serialize_source(source)
    fingerprint = hashlib.sha256(serialized.encode(encoding="utf-8"))
    fingerprint_hex_str = fingerprint.hexdigest()
    return source.entry_point.name + "_" + fingerprint_hex_str


def get_cache_folder(
    compilable_source: stages.CompilableSource, lifetime: config.BuildCacheLifetime
) -> pathlib.Path:
    """
    Construct the path to where the build system project artifact of a compilable source should be cached.

    The returned path points to an existing folder in all cases.
    """
    # TODO(ricoh): make dependent on binding source too or add alternative that depends on bindings
    folder_name = _cache_folder_name(compilable_source.program_source)

    match lifetime:
        case config.BuildCacheLifetime.SESSION:
            base_path = _session_cache_dir_path
        case config.BuildCacheLifetime.PERSISTENT:
            base_path = config.BUILD_CACHE_DIR
        case _:
            raise ValueError("Unsupported caching lifetime.")

    base_path.mkdir(exist_ok=True)

    complete_path = base_path / folder_name
    complete_path.mkdir(exist_ok=True)

    # Resolve symlinks to workaround an issue on MacOS where the default tmp directory is a symlink
    # which might sometimes get resolved in a way we don't control.
    return complete_path.resolve()



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Implements the lowering of scan field operator.

This builtin translator implements the `PrimitiveTranslator` protocol as other
translators in `gtir_to_sdfg_primitives` module. This module implements the scan
translator, separately from the `gtir_to_sdfg_primitives` module, because the
parsing of input arguments as well as the construction of the map scope differ
from a regular field operator, which requires slightly different helper methods.
Besides, the function code is quite large, another reason to keep it separate
from other translators.

The current GTIR representation of the scan operator is based on iterator view.
This is likely to change in the future, to enable GTIR optimizations for scan.
"""

from __future__ import annotations

import itertools
from typing import Any, Iterable

import dace
from dace import subsets as dace_subsets

from gt4py import eve
from gt4py.next import common as gtx_common, utils as gtx_utils
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.program_processors.runners.dace import (
    gtir_dataflow,
    gtir_domain,
    gtir_to_sdfg,
    gtir_to_sdfg_types,
    gtir_to_sdfg_utils,
)
from gt4py.next.type_system import type_info as ti, type_specifications as ts


def _parse_scan_fieldop_arg(
    node: gtir.Expr,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    domain: gtir_domain.FieldopDomain,
) -> gtir_dataflow.MemletExpr | tuple[gtir_dataflow.MemletExpr | tuple[Any, ...], ...]:
    """Helper method to visit an expression passed as argument to a scan field operator.

    On the innermost level, a scan operator is lowered to a loop region which computes
    column elements in the vertical dimension.

    It differs from the helper method `gtir_to_sdfg_primitives` in that field arguments
    are passed in full shape along the vertical dimension, rather than as iterator.
    """

    def _parse_fieldop_arg_impl(
        arg: gtir_to_sdfg_types.FieldopData,
    ) -> gtir_dataflow.MemletExpr:
        arg_expr = arg.get_local_view(domain, ctx.sdfg)
        if isinstance(arg_expr, gtir_dataflow.MemletExpr):
            return arg_expr
        # In scan field operator, the arguments to the vertical stencil are passed by value.
        # Therefore, the full field shape is passed as `MemletExpr` rather than `IteratorExpr`.
        return gtir_dataflow.MemletExpr(
            arg_expr.field, arg_expr.gt_dtype, arg_expr.get_memlet_subset(ctx.sdfg)
        )

    arg = sdfg_builder.visit(node, ctx=ctx)

    if isinstance(arg, gtir_to_sdfg_types.FieldopData):
        return _parse_fieldop_arg_impl(arg)
    else:
        # handle tuples of fields
        return gtx_utils.tree_map(lambda x: _parse_fieldop_arg_impl(x))(arg)


def _create_scan_field_operator_impl(
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    field_domain: gtir_domain.FieldopDomain,
    output_edge: gtir_dataflow.DataflowOutputEdge,
    output_type: ts.FieldType,
    map_exit: dace.nodes.MapExit,
) -> gtir_to_sdfg_types.FieldopData:
    """
    Helper method to allocate a temporary array that stores one field computed
    by the scan field operator.

    This method is called by `_create_scan_field_operator()`.

    Similar to `gtir_to_sdfg_primitives._create_field_operator_impl()` but
    for scan field operators. It differs in that the scan loop region produces
    a field along the vertical dimension, rather than a single point.
    Therefore, the memlet subset will write a slice into the result array, that
    corresponds to the full vertical shape for each horizontal grid point.

    Refer to `gtir_to_sdfg_primitives._create_field_operator_impl()` for
    the description of function arguments and return values.
    """
    dataflow_output_desc = output_edge.result.dc_node.desc(ctx.sdfg)
    assert isinstance(dataflow_output_desc, dace.data.Array)

    # the memory layout of the output field follows the field operator compute domain
    field_dims, field_origin, field_shape = gtir_domain.get_field_layout(
        field_domain, ctx.target_domain
    )
    field_indices = gtir_domain.get_domain_indices(field_dims, field_origin)
    field_subset = dace_subsets.Range.from_indices(field_indices)

    # the vertical dimension used as scan column is computed by the `LoopRegion`
    # inside the map scope, therefore it is excluded from the map range
    scan_dim_index = [sdfg_builder.is_column_axis(dim) for dim in field_dims].index(True)

    # the map scope writes the full-shape dimension corresponding to the scan column
    field_subset = (
        dace_subsets.Range(field_subset[:scan_dim_index])
        + dace_subsets.Range.from_string(f"0:{dataflow_output_desc.shape[0]}")
        + dace_subsets.Range(field_subset[scan_dim_index + 1 :])
    )

    if isinstance(output_edge.result.gt_dtype, ts.ScalarType):
        assert isinstance(output_type.dtype, ts.ScalarType)
        if output_edge.result.gt_dtype != output_type.dtype:
            raise TypeError(
                f"Type mismatch, expected {output_type.dtype} got {output_edge.result.gt_dtype}."
            )
        field_dtype = output_edge.result.gt_dtype
        # the scan field operator computes a column of scalar values
        assert len(dataflow_output_desc.shape) == 1
    else:
        assert isinstance(output_type.dtype, ts.ListType)
        assert isinstance(output_edge.result.gt_dtype.element_type, ts.ScalarType)
        field_dtype = output_edge.result.gt_dtype.element_type
        if field_dtype != output_type.dtype.element_type:
            raise TypeError(
                f"Type mismatch, expected {output_type.dtype.element_type} got {field_dtype}."
            )
        # the scan field operator computes a list of scalar values for each column level
        # 1st dim: column level, 2nd dim: list of scalar values (e.g. `neighbors`)
        assert len(dataflow_output_desc.shape) == 2
        # the lines below extend the array with the local dimension added by the field operator
        assert output_edge.result.gt_dtype.offset_type is not None
        field_shape = [*field_shape, dataflow_output_desc.shape[1]]
        field_subset = field_subset + dace_subsets.Range.from_string(
            f"0:{dataflow_output_desc.shape[1]}"
        )

    # allocate local temporary storage
    field_name, field_desc = sdfg_builder.add_temp_array(
        ctx.sdfg, field_shape, dataflow_output_desc.dtype
    )
    # the inner and outer strides have to match
    scan_output_stride = field_desc.strides[scan_dim_index]
    # also consider the stride of the local dimension, in case the scan field operator computes a list
    local_strides = field_desc.strides[len(field_dims) :]
    assert len(local_strides) == (1 if isinstance(output_edge.result.gt_dtype, ts.ListType) else 0)
    new_inner_strides = [scan_output_stride, *local_strides]
    dataflow_output_desc.set_shape(dataflow_output_desc.shape, new_inner_strides)

    # and here the edge writing the dataflow result data through the map exit node
    field_node = ctx.state.add_access(field_name)
    output_edge.connect(map_exit, field_node, field_subset)

    return gtir_to_sdfg_types.FieldopData(
        field_node, ts.FieldType(field_dims, output_edge.result.gt_dtype), tuple(field_origin)
    )


def _create_scan_field_operator(
    ctx: gtir_to_sdfg.SubgraphContext,
    domain: gtir_domain.FieldopDomain,
    node_type: ts.FieldType | ts.TupleType,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    input_edges: Iterable[gtir_dataflow.DataflowInputEdge],
    output_tree: gtir_dataflow.DataflowOutputEdge
    | tuple[gtir_dataflow.DataflowOutputEdge | tuple[Any, ...], ...],
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Helper method to build the output of a field operator, which can consist of
    a single field or a tuple of fields.

    Similar to `gtir_to_sdfg_primitives._create_field_operator()` but for scan
    field operators. The main difference is that the scan vertical dimension is
    excluded from the map range. This because the vertical dimension is traversed
    by a loop region in a mapped nested SDFG.

    Refer to `gtir_to_sdfg_primitives._create_field_operator()` for the
    description of function arguments and return values.
    """
    domain_dims, _, _ = gtir_domain.get_field_layout(domain, ctx.target_domain)

    # create a map scope to execute the `LoopRegion` over the horizontal domain
    if len(domain_dims) == 1:
        # We construct the scan field operator on the horizontal domain, while the
        # vertical dimension (the column axis) is computed by the loop region.
        # If the field operator computes only the column axis (a 1d scan field operator),
        # there is no horizontal domain, therefore the map scope is not needed.
        # This case currently produces wrong CUDA code because of a DaCe issue
        # (see https://github.com/GridTools/gt4py/issues/1136).
        # The corresponding GT4Py tests are disabled (pytest marker `uses_scan_1d_field`).
        map_entry, map_exit = (None, None)
    else:
        # create map range corresponding to the field operator domain
        map_entry, map_exit = sdfg_builder.add_map(
            "fieldop",
            ctx.state,
            ndrange={
                gtir_to_sdfg_utils.get_map_variable(
                    domain_range.dim
                ): f"{domain_range.start}:{domain_range.stop}"
                for domain_range in domain
                if not sdfg_builder.is_column_axis(domain_range.dim)
            },
        )

    # here we setup the edges passing through the map entry node
    for edge in input_edges:
        edge.connect(map_entry)

    if isinstance(node_type, ts.FieldType):
        assert isinstance(output_tree, gtir_dataflow.DataflowOutputEdge)
        return _create_scan_field_operator_impl(
            ctx, sdfg_builder, domain, output_tree, node_type, map_exit
        )
    else:
        # handle tuples of fields
        # the symbol name 'x' in the call below is not used, we only need
        # the tree structure of the `TupleType` definition to pass to `tree_map()`
        output_symbol_tree = gtir_to_sdfg_utils.make_symbol_tree("x", node_type)
        return gtx_utils.tree_map(
            lambda output_edge, output_sym: (
                _create_scan_field_operator_impl(
                    ctx,
                    sdfg_builder,
                    domain,
                    output_edge,
                    output_sym.type,
                    map_exit,
                )
            )
        )(output_tree, output_symbol_tree)


def _scan_input_name(input_name: str) -> str:
    """
    Helper function to make naming of input connectors in the scan nested SDFG
    consistent throughut this module scope.
    """
    return f"__gtir_scan_input_{input_name}"


def _scan_output_name(input_name: str) -> str:
    """
    Same as above, but for the output connecters in the scan nested SDFG.
    """
    return f"__gtir_scan_output_{input_name}"


def _lower_lambda_to_nested_sdfg(
    lambda_node: gtir.Lambda,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    domain: gtir_domain.FieldopDomain,
    init_data: gtir_to_sdfg_types.FieldopResult,
    lambda_symbols: dict[str, ts.DataType],
    scan_forward: bool,
    scan_carry_symbol: gtir.Sym,
) -> tuple[gtir_to_sdfg.SubgraphContext, gtir_to_sdfg_types.FieldopResult]:
    """
    Helper method to lower the lambda node representing the scan stencil dataflow
    inside a separate SDFG.

    In regular field operators, where the computation of a grid point is independent
    from other points, therefore the stencil can be lowered to a mapped tasklet
    dataflow, and the map range is defined on the full domain.
    The scan field operator has to carry an intermediate result while the stencil
    is applied on vertical levels, which is input to the computation of next level
    (an accumulator function, for example). Therefore, the points on the vertical
    dimension are computed inside a `LoopRegion` construct.
    This function creates the `LoopRegion` inside a nested SDFG, which will be
    mapped by the caller to the horizontal domain in the field operator context.

    Args:
        lambda_node: The lambda representing the stencil expression on the horizontal level.
        ctx: The SDFG context where the scan field operator is translated.
        sdfg_builder: The SDFG builder object to access the field operator context.
        domain: The field operator domain, with all horizontal and vertical dimensions.
        init_data: The data produced in the field operator context that is used
            to initialize the scan carry value.
        lambda_symbols: List of symbols used as parameters of the stencil expressions.
        scan_forward: When True, the loop should range starting from the origin;
            when False, traverse towards origin.
        scan_carry_symbol: The symbol used in the stencil expression to carry the
            intermediate result along the vertical dimension.

    Returns:
        A tuple of two elements:
          - The subgraph context containing the `LoopRegion` along the vertical
            dimension, to be instantied as a nested SDFG in the field operator context.
          - The inner fields, that is 1d arrays with vertical shape containing
            the output of the stencil computation. These fields will have to be
            mapped to outer arrays by the caller. The caller is responsible to ensure
            that inner and outer arrays use the same strides.
    """

    # We pass an empty set as symbolic arguments, which implies that all scalar
    # inputs of the scan nested SDFG will be represented as scalar data containers.
    # The reason why we do not check for dace symbolic expressions and do not map
    # them to inner symbols is that the scan expression should not contain any domain
    # expression (no field operator inside).
    assert not any(
        eve.walk_values(lambda_node).map(
            lambda node: cpm.is_call_to(node, ("cartesian_domain", "unstructured_domain"))
        )
    )
    # the lambda expression, i.e. body of the scan, will be created inside a nested SDFG.
    lambda_translator, lambda_ctx = sdfg_builder.setup_nested_context(
        lambda_node, "scan", ctx, lambda_symbols, symbolic_inputs=set()
    )

    # We set `using_explicit_control_flow=True` because the vertical scan is lowered to a `LoopRegion`.
    # This property is used by pattern matching in SDFG transformation framework
    # to skip those transformations that do not yet support control flow blocks.
    lambda_ctx.sdfg.using_explicit_control_flow = True

    # use the vertical dimension in the domain as scan dimension
    scan_domain = next(
        domain_range for domain_range in domain if sdfg_builder.is_column_axis(domain_range.dim)
    )

    # extract the scan loop range
    scan_loop_var = gtir_to_sdfg_utils.get_map_variable(scan_domain.dim)

    # in case the scan operator computes a list (not a scalar), we need to add an extra dimension
    def get_scan_output_shape(
        scan_init_data: gtir_to_sdfg_types.FieldopData,
    ) -> list[dace.symbolic.SymExpr]:
        scan_column_size = scan_domain.stop - scan_domain.start
        if isinstance(scan_init_data.gt_type, ts.ScalarType):
            return [scan_column_size]
        assert isinstance(scan_init_data.gt_type, ts.ListType)
        assert scan_init_data.gt_type.offset_type
        offset_type = scan_init_data.gt_type.offset_type
        offset_provider_type = sdfg_builder.get_offset_provider_type(offset_type.value)
        assert isinstance(offset_provider_type, gtx_common.NeighborConnectivityType)
        list_size = offset_provider_type.max_neighbors
        return [scan_column_size, dace.symbolic.SymExpr(list_size)]

    if isinstance(init_data, tuple):
        lambda_result_shape = gtx_utils.tree_map(get_scan_output_shape)(init_data)
    else:
        lambda_result_shape = get_scan_output_shape(init_data)

    # Create the body of the initialization state
    # This dataflow will write the initial value of the scan carry variable.
    init_state = lambda_ctx.state
    scan_carry_input = (
        gtir_to_sdfg_utils.make_symbol_tree(scan_carry_symbol.id, scan_carry_symbol.type)
        if isinstance(scan_carry_symbol.type, ts.TupleType)
        else scan_carry_symbol
    )

    def init_scan_carry(sym: gtir.Sym) -> None:
        scan_carry_dataname = str(sym.id)
        scan_carry_desc = lambda_ctx.sdfg.data(scan_carry_dataname)
        input_scan_carry_dataname = _scan_input_name(scan_carry_dataname)
        input_scan_carry_desc = scan_carry_desc.clone()
        lambda_ctx.sdfg.add_datadesc(input_scan_carry_dataname, input_scan_carry_desc)
        scan_carry_desc.transient = True
        init_state.add_nedge(
            init_state.add_access(input_scan_carry_dataname),
            init_state.add_access(scan_carry_dataname),
            lambda_ctx.sdfg.make_array_memlet(input_scan_carry_dataname),
        )

    if isinstance(scan_carry_input, tuple):
        gtx_utils.tree_map(init_scan_carry)(scan_carry_input)
    else:
        init_scan_carry(scan_carry_input)

    # Create a loop region over the vertical dimension corresponding to the scan column
    if scan_forward:
        scan_loop = dace.sdfg.state.LoopRegion(
            label="scan",
            condition_expr=f"{scan_loop_var} < {scan_domain.stop}",
            loop_var=scan_loop_var,
            initialize_expr=f"{scan_loop_var} = {scan_domain.start}",
            update_expr=f"{scan_loop_var} = {scan_loop_var} + 1",
            inverted=False,
        )
    else:
        scan_loop = dace.sdfg.state.LoopRegion(
            label="scan",
            condition_expr=f"{scan_loop_var} >= {scan_domain.start}",
            loop_var=scan_loop_var,
            initialize_expr=f"{scan_loop_var} = {scan_domain.stop} - 1",
            update_expr=f"{scan_loop_var} = {scan_loop_var} - 1",
            inverted=False,
        )
    lambda_ctx.sdfg.add_node(scan_loop)
    lambda_ctx.sdfg.add_edge(init_state, scan_loop, dace.InterstateEdge())

    # Inside the loop region, create a 'compute' and an 'update' state.
    # The body of the 'compute' state implements the stencil expression for one vertical level.
    # The 'update' state writes the value computed by the stencil into the scan carry variable,
    # in order to make it available to the next vertical level.
    compute_state = scan_loop.add_state("scan_compute")
    compute_ctx = gtir_to_sdfg.SubgraphContext(
        lambda_ctx.sdfg, compute_state, lambda_ctx.target_domain
    )
    update_state = scan_loop.add_state_after(compute_state, "scan_update")

    # inside the 'compute' state, visit the list of arguments to be passed to the stencil
    stencil_args = [
        _parse_scan_fieldop_arg(im.ref(p.id), compute_ctx, lambda_translator, domain)
        for p in lambda_node.params
    ]
    # stil inside the 'compute' state, generate the dataflow representing the stencil
    # to be applied on the horizontal domain
    lambda_input_edges, lambda_result = gtir_dataflow.translate_lambda_to_dataflow(
        compute_ctx.sdfg, compute_ctx.state, lambda_translator, lambda_node, stencil_args
    )
    # connect the dataflow input directly to the source data nodes, without passing through a map node;
    # the reason is that the map for horizontal domain is outside the scan loop region
    for edge in lambda_input_edges:
        edge.connect(map_entry=None)
    # connect the dataflow output nodes, called 'scan_result' below, to a global field called 'output'
    output_column_index = dace.symbolic.pystr_to_symbolic(scan_loop_var) - scan_domain.start

    def connect_scan_output(
        scan_output_edge: gtir_dataflow.DataflowOutputEdge,
        scan_output_shape: list[dace.symbolic.SymExpr],
        scan_carry_sym: gtir.Sym,
    ) -> gtir_to_sdfg_types.FieldopData:
        scan_result = scan_output_edge.result
        if isinstance(scan_result.gt_dtype, ts.ScalarType):
            assert scan_result.gt_dtype == scan_carry_sym.type
            # the scan field operator computes a column of scalar values
            assert len(scan_output_shape) == 1
            output_subset = dace_subsets.Range.from_string(str(output_column_index))
        else:
            assert isinstance(scan_carry_sym.type, ts.ListType)
            assert scan_result.gt_dtype.element_type == scan_carry_sym.type.element_type
            # the scan field operator computes a list of scalar values for each column level
            assert len(scan_output_shape) == 2
            output_subset = dace_subsets.Range.from_string(
                f"{output_column_index}, 0:{scan_output_shape[1]}"
            )
        scan_result_data = scan_result.dc_node.data
        scan_result_desc = scan_result.dc_node.desc(lambda_ctx.sdfg)
        scan_result_subset = dace_subsets.Range.from_array(scan_result_desc)

        # `sym` represents the global output data, that is the nested-SDFG output connector
        scan_carry_data = str(scan_carry_sym.id)
        output = _scan_output_name(scan_carry_data)
        lambda_ctx.sdfg.add_array(output, scan_output_shape, scan_result_desc.dtype)
        output_node = compute_state.add_access(output)

        # in the 'compute' state, we write the current vertical level data to the output field
        # (the output field is mapped to an external array)
        compute_state.add_nedge(
            scan_result.dc_node,
            output_node,
            dace.Memlet(data=output, subset=output_subset, other_subset=scan_result_subset),
        )

        # in the 'update' state, the value of the current vertical level is written
        # to the scan carry variable for the next loop iteration
        update_state.add_nedge(
            update_state.add_access(scan_result_data),
            update_state.add_access(scan_carry_data),
            dace.Memlet(
                data=scan_result_data, subset=scan_result_subset, other_subset=scan_result_subset
            ),
        )

        output_type = ts.FieldType(dims=[scan_domain.dim], dtype=scan_result.gt_dtype)
        return gtir_to_sdfg_types.FieldopData(output_node, output_type, origin=(scan_domain.start,))

    # write the stencil result (value on one vertical level) into a 1D field
    # with full vertical shape representing one column
    if isinstance(scan_carry_input, tuple):
        assert isinstance(lambda_result_shape, tuple)
        lambda_output = gtx_utils.tree_map(connect_scan_output)(
            lambda_result, lambda_result_shape, scan_carry_input
        )
    else:
        assert isinstance(lambda_result[0], gtir_dataflow.DataflowOutputEdge)
        assert isinstance(lambda_result_shape, list)
        lambda_output = connect_scan_output(lambda_result[0], lambda_result_shape, scan_carry_input)

    # in case tuples are passed as argument, isolated access nodes might be left in the state,
    # because not all tuple fields are necessarily accessed inside the lambda scope
    for data_node in compute_state.data_nodes():
        if compute_state.degree(data_node) == 0:
            # By construction there should never be isolated transient nodes.
            # Therefore, the assert below implements a sanity check, that allows
            # the exceptional case (encountered in one GT4Py test) where the carry
            # variable is not used, so not a scan indeed because no data dependency.
            assert (not data_node.desc(lambda_ctx.sdfg).transient) or data_node.data.startswith(
                scan_carry_symbol.id
            )
            compute_state.remove_node(data_node)

    return lambda_ctx, lambda_output


def _connect_nested_sdfg_output_to_temporaries(
    inner_ctx: gtir_to_sdfg.SubgraphContext,
    outer_ctx: gtir_to_sdfg.SubgraphContext,
    nsdfg_node: dace.nodes.NestedSDFG,
    inner_data: gtir_to_sdfg_types.FieldopData,
) -> gtir_dataflow.DataflowOutputEdge:
    """
    Helper function to create the edges to write output data from the nested SDFG
    to temporary arrays in the parent SDFG, denoted as outer context.

    Args:
        inner_ctx: The inner SDFG context, where the scan `LoopRegion` is translated.
        outer_ctx: The outer SDFG context, where the field operator is translated.
        nsdfg_node: The nested SDFG node in the outer context.
        inner_data: The data produced by the scan `LoopRegion` in the inner context.

    Returns:
        An object representing the output data connection of this field operator.
    """
    assert isinstance(inner_data.gt_type, ts.FieldType)
    inner_dataname = inner_data.dc_node.data
    inner_desc = inner_ctx.sdfg.data(inner_dataname)
    outer_dataname, outer_desc = outer_ctx.sdfg.add_temp_transient_like(inner_desc)
    outer_node = outer_ctx.state.add_access(outer_dataname)
    outer_ctx.state.add_edge(
        nsdfg_node,
        inner_dataname,
        outer_node,
        None,
        dace.Memlet.from_array(outer_dataname, outer_desc),
    )
    output_expr = gtir_dataflow.ValueExpr(outer_node, inner_data.gt_type.dtype)
    return gtir_dataflow.DataflowOutputEdge(outer_ctx.state, output_expr)


def translate_scan(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Generates the dataflow subgraph for the `as_fieldop` builtin with a scan operator.

    It differs from `translate_as_fieldop()` in that the horizontal domain is lowered
    to a map scope, while the scan column computation is lowered to a `LoopRegion`
    on the vertical dimension, that is inside the horizontal map.
    The current design choice is to keep the map scope on the outer level, and
    the `LoopRegion` inside. This choice follows the GTIR representation where
    the `scan` operator is called inside the `as_fieldop` node.

    Implements the `PrimitiveTranslator` protocol.
    """
    assert isinstance(node, gtir.FunCall)
    assert cpm.is_call_to(node.fun, "as_fieldop")
    assert isinstance(node.type, (ts.FieldType, ts.TupleType))

    fun_node = node.fun
    assert len(fun_node.args) == 2
    scan_expr, domain_expr = fun_node.args
    assert cpm.is_call_to(scan_expr, "scan")

    # parse the domain of the scan field operator
    domain = gtir_domain.extract_domain(domain_expr)

    # parse scan parameters
    assert len(scan_expr.args) == 3
    stencil_expr = scan_expr.args[0]
    assert isinstance(stencil_expr, gtir.Lambda)

    # params[0]: the lambda parameter to propagate the scan carry on the vertical dimension
    scan_carry = str(stencil_expr.params[0].id)

    # params[1]: boolean flag for forward/backward scan
    assert isinstance(scan_expr.args[1], gtir.Literal) and ti.is_logical(scan_expr.args[1].type)
    scan_forward = scan_expr.args[1].value == "True"

    # params[2]: the expression that computes the value for scan initialization
    init_expr = scan_expr.args[2]
    # visit the initialization value of the scan expression
    init_data = sdfg_builder.visit(init_expr, ctx=ctx)
    # extract type definition of the scan carry
    scan_carry_type = (
        init_data.gt_type
        if isinstance(init_data, gtir_to_sdfg_types.FieldopData)
        else gtir_to_sdfg_types.get_tuple_type(init_data)
    )

    # define the set of symbols available in the lambda context, which consists of
    # the carry argument and all lambda function arguments
    lambda_arg_types = [scan_carry_type] + [
        arg.type for arg in node.args if isinstance(arg.type, ts.DataType)
    ]
    lambda_symbols = {
        str(p.id): arg_type
        for p, arg_type in zip(stencil_expr.params, lambda_arg_types, strict=True)
    }

    # lower the scan stencil expression in a separate SDFG context
    lambda_ctx, lambda_output = _lower_lambda_to_nested_sdfg(
        stencil_expr,
        ctx,
        sdfg_builder,
        domain,
        init_data,
        lambda_symbols,
        scan_forward,
        im.sym(scan_carry, scan_carry_type),
    )

    # visit the arguments to be passed to the lambda expression
    # this must be executed before visiting the lambda expression, in order to populate
    # the data descriptor with the correct field domain offsets for field arguments
    lambda_args = [sdfg_builder.visit(arg, ctx=ctx) for arg in node.args]
    lambda_args_mapping = [
        (im.sym(_scan_input_name(scan_carry), scan_carry_type), init_data),
    ] + [
        (im.sym(param.id, arg.gt_type), arg)
        for param, arg in zip(stencil_expr.params[1:], lambda_args, strict=True)
    ]

    lambda_arg_nodes = dict(
        itertools.chain(
            *[gtir_to_sdfg_types.flatten_tuples(psym.id, arg) for psym, arg in lambda_args_mapping]
        )
    )

    # parse the dataflow output symbols
    if isinstance(scan_carry_type, ts.TupleType):
        lambda_flat_outs = {
            str(sym.id): sym.type
            for sym in gtir_to_sdfg_utils.flatten_tuple_fields(
                _scan_output_name(scan_carry), scan_carry_type
            )
        }
    else:
        lambda_flat_outs = {_scan_output_name(scan_carry): scan_carry_type}

    # build the mapping of symbols from nested SDFG to field operator context
    nsdfg_symbols_mapping = {str(sym): sym for sym in lambda_ctx.sdfg.free_symbols}
    for psym, arg in lambda_args_mapping:
        nsdfg_symbols_mapping |= gtir_to_sdfg_utils.get_arg_symbol_mapping(psym.id, arg, ctx.sdfg)

    # the scan nested SDFG is ready: it is instantiated in the field operator context
    # where the map scope over the horizontal domain lives
    nsdfg_node = ctx.state.add_nested_sdfg(
        lambda_ctx.sdfg,
        inputs=set(lambda_arg_nodes.keys()),
        outputs=set(lambda_flat_outs.keys()),
        symbol_mapping=nsdfg_symbols_mapping,
    )

    lambda_input_edges = []
    for input_connector, outer_arg in lambda_arg_nodes.items():
        arg_desc = outer_arg.dc_node.desc(ctx.sdfg)
        input_subset = dace_subsets.Range.from_array(arg_desc)
        input_edge = gtir_dataflow.MemletInputEdge(
            ctx.state, outer_arg.dc_node, input_subset, nsdfg_node, input_connector
        )
        lambda_input_edges.append(input_edge)

    # for output connections, we create temporary arrays that contain the computation
    # results of a column slice for each point in the horizontal domain
    lambda_output_tree = gtx_utils.tree_map(
        lambda lambda_output_data: _connect_nested_sdfg_output_to_temporaries(
            lambda_ctx, ctx, nsdfg_node, lambda_output_data
        )
    )(lambda_output)

    # we call a helper method to create a map scope that will compute the entire field
    return _create_scan_field_operator(
        ctx, domain, node.type, sdfg_builder, lambda_input_edges, lambda_output_tree
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Iterator, Optional, Sequence, Union

from gt4py.eve import datamodels as eve_datamodels, type_definitions as eve_types
from gt4py.next import common


class TypeSpec(eve_datamodels.DataModel, kw_only=False, frozen=True): ...  # type: ignore[call-arg]


class DataType(TypeSpec):
    """
    A base type for all types that represent data storage.

    Derive floating point, integral or field types from this class.
    """


class CallableType:
    """
    A base type for all types are callable.

    Derive other callable types, such as FunctionType or FieldOperatorType from
    this class.
    """


class DeferredType(TypeSpec):
    """Dummy used to represent a type not yet inferred."""

    constraint: Optional[type[TypeSpec] | tuple[type[TypeSpec], ...]]


class VoidType(TypeSpec):
    """
    Return type of a function without return values.

    Note: only useful for stateful dialects.
    """


class DimensionType(TypeSpec):
    dim: common.Dimension

    def __str__(self) -> str:
        return str(self.dim)


class IndexType(TypeSpec):
    """
    Represents the type of an index into a dimension.
    """

    dim: common.Dimension

    def __str__(self) -> str:
        return f"Index[{self.dim}]"


class OffsetType(TypeSpec):
    # TODO(havogt): replace by ConnectivityType
    source: common.Dimension
    target: tuple[common.Dimension] | tuple[common.Dimension, common.Dimension]

    def __str__(self) -> str:
        return f"Offset[{self.source}, {self.target}]"


class ScalarKind(eve_types.IntEnum):
    BOOL = 1
    INT8 = 2
    UINT8 = 3
    INT16 = 4
    UINT16 = 5
    INT32 = 6
    UINT32 = 7
    INT64 = 8
    UINT64 = 9
    FLOAT32 = 10
    FLOAT64 = 11
    STRING = 12


class ScalarType(DataType):
    kind: ScalarKind
    shape: Optional[list[int]] = None

    def __str__(self) -> str:
        kind_str = self.kind.name.lower()
        if self.shape is None:
            return kind_str
        return f"{kind_str}{self.shape}"


class ListType(DataType):
    """Represents a neighbor list in the ITIR representation.

    Note: not used in the frontend. The concept is represented as Field with local Dimension.
    """

    element_type: DataType
    # TODO(tehrengruber): make `offset_type` mandatory
    offset_type: Optional[common.Dimension] = None


class FieldType(DataType, CallableType):
    dims: list[common.Dimension]
    dtype: ScalarType | ListType

    def __str__(self) -> str:
        dims = "..." if self.dims is Ellipsis else f"[{', '.join(dim.value for dim in self.dims)}]"
        return f"Field[{dims}, {self.dtype}]"

    @eve_datamodels.validator("dims")
    def _dims_validator(
        self, attribute: eve_datamodels.Attribute, dims: list[common.Dimension]
    ) -> None:
        common.check_dims(dims)


class TupleType(DataType):
    # TODO(tehrengruber): Remove `DeferredType` again. This was erroneously
    #  introduced before we checked the annotations at runtime. All attributes of
    #  a type that are types themselves must be concrete.
    types: list[DataType | DimensionType | DeferredType]

    def __str__(self) -> str:
        return f"tuple[{', '.join(map(str, self.types))}]"

    def __iter__(self) -> Iterator[DataType | DimensionType | DeferredType]:
        yield from self.types

    def __len__(self) -> int:
        return len(self.types)


class FunctionType(TypeSpec, CallableType):
    pos_only_args: Sequence[TypeSpec]
    pos_or_kw_args: dict[str, TypeSpec]
    kw_only_args: dict[str, TypeSpec]
    returns: Union[TypeSpec]

    def __str__(self) -> str:
        arg_strs = [str(arg) for arg in self.pos_only_args]
        kwarg_strs = [f"{key}: {value}" for key, value in self.pos_or_kw_args.items()]
        args_str = ", ".join((*arg_strs, *kwarg_strs))
        return f"({args_str}) -> {self.returns}"


class DomainType(DataType):
    dims: list[common.Dimension]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from . import gtscript_frontend
from .base import REGISTRY, Frontend, from_name, register


__all__ = ["REGISTRY", "Frontend", "from_name", "gtscript_frontend", "register"]
